###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCb Bookkeeping database manager."""
import os
import time

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.BookkeepingSystem.Client import objects
from LHCbDIRAC.BookkeepingSystem.Client.Help import helpConfig, helpProcessing, helpEventType

INTERNAL_PATH_SEPARATOR = "/"

#############################################################################


class LHCbBookkeepingManager:
    """creates the virtual file system."""

    __bookkeepingFolderProperties = [
        "name",
        "fullpath",
    ]
    # watch out for this ad hoc solution
    # if any changes made check all functions
    #
    __bookkeepingConfigurationPrefixes = [
        "ConfigName",  # configname
        "ConfigVersion",  # configversion
        "Simulation/DataTaking",
        "ProcessingPass",
        "EventType",  # event type
        "Production",  # production
        "FileType",  # file type
        "",
    ]

    __bookkeepingProductionPrefixes = ["PROD", "EVT", "FTY", ""]

    __bookkeepingRunPrefixes = ["RUN", "PAS", "EVT", "FTY", ""]

    __bookkeepingEventtypePrefixes = [
        "ConfigName",
        "ConfigVersion",
        "EventType",
        "Simulation/DataTaking",
        "ProcessingPass",
        "Production",
        "FileType",
        "",
    ]
    __bookkeepingDatabasePrefixes = []

    __bookkeepingParameters = ["Configuration", "Event type", "Productions", "Runlookup"]

    __bookkeepingShortparameternames = {
        __bookkeepingParameters[0]: "sim",
        __bookkeepingParameters[1]: "evt",
        __bookkeepingParameters[2]: "prod",
        __bookkeepingParameters[3]: "run",
    }

    __bookkeepingQueryTypes = ["adv", "std"]

    #############################################################################
    def __init__(self):
        """initialize the values."""
        self.db_ = BookkeepingClient()
        self.fileCatalog = FileCatalog()

        self.__entityCache = {"/": (objects.Entity({"name": "/", "fullpath": "/", "expandable": True}), 0)}
        self.parameter_ = self.__bookkeepingParameters[0]
        self.files_ = []
        self.__filetypes = []
        self.comment = "#-- "

        self.treeLevels_ = -1
        self.advancedQuery_ = False
        self.dataQualities_ = {}

        retVal = self.db_.getAvailableFileTypes()
        if not retVal["OK"]:
            gLogger.error(retVal["Message"])
        else:
            self.__filetypes = [i[0] for i in retVal["Value"]["Records"]]

    def setFileTypes(self, fileTypeList=list()):
        """it sets the file types.

        The parameter is a list of file type
        """
        if fileTypeList and len(fileTypeList) > 0:
            self.__filetypes = fileTypeList
        else:
            retVal = self.db_.getAvailableFileTypes()
            if not retVal["OK"]:
                gLogger.error(retVal)
            else:
                self.__filetypes = [i[0] for i in retVal["Value"]["Records"]]

    #############################################################################
    def setAdvancedQueries(self, value):
        """advanced queries."""
        self.advancedQuery_ = value

    #############################################################################
    def help(self):
        """help information."""
        if self.parameter_ == self.__bookkeepingParameters[0]:
            helpConfig(self.treeLevels_)
        elif self.parameter_ == self.__bookkeepingParameters[1]:
            helpEventType(self.treeLevels_)
        elif self.parameter_ == self.__bookkeepingParameters[2]:
            helpProcessing(self.treeLevels_)

    #############################################################################
    def getPossibleParameters(self):
        """available parameters."""
        return self.__bookkeepingParameters

    #############################################################################
    def getCurrentParameter(self):
        """current parameters."""
        return self.__bookkeepingShortparameternames[self.parameter_]

    #############################################################################
    def getQueriesTypes(self):
        """types of queries."""
        if self.advancedQuery_:
            return self.__bookkeepingQueryTypes[0]
        else:
            return self.__bookkeepingQueryTypes[1]

    #############################################################################
    def setParameter(self, name):
        """query types."""
        if self.__bookkeepingParameters.__contains__(name):
            self.parameter_ = name
            self.treeLevels_ = -1
            if name == "Configuration":
                self.__bookkeepingDatabasePrefixes = self.__bookkeepingConfigurationPrefixes
            elif name == "Productions":
                self.__bookkeepingDatabasePrefixes = self.__bookkeepingProductionPrefixes
            elif name == "Event type":
                self.__bookkeepingDatabasePrefixes = self.__bookkeepingEventtypePrefixes
            elif name == "Runlookup":
                self.__bookkeepingDatabasePrefixes = self.__bookkeepingRunPrefixes
        else:
            gLogger.error("Wrong Parameter!")

    #############################################################################
    def getFilesPFN(self):
        """pfn."""
        return self.fileCatalog.getReplicas(self.files_)

    #############################################################################
    def list(self, path="/", selectionDict=None, sortDict=None, startItem=0, maxitems=0):
        """list a path."""
        gLogger.debug(path)
        selectionDict = selectionDict if selectionDict is not None else {}
        sortDict = sortDict if sortDict is not None else {}

        if self.parameter_ == self.__bookkeepingParameters[0]:
            return self._listConfigs(path, sortDict, startItem, maxitems)
        elif self.parameter_ == self.__bookkeepingParameters[1]:
            return self._listEventTypes(path, sortDict, startItem, maxitems)
        elif self.parameter_ == self.__bookkeepingParameters[2]:
            return self._listProduction(path)
        elif self.parameter_ == self.__bookkeepingParameters[3]:
            return self._listRuns(path)

    #############################################################################
    def getLevelAndPath(self, path):
        """level and path."""
        if path == "/":
            return 0, [], ""  # it is the first level
        path = self.getAbsolutePath(path)["Value"]  # shall we do this here or in the _processedPath()?
        processedPath = self._processPath(path)
        tmpPath = list(processedPath)
        if self.parameter_ == self.__bookkeepingParameters[1]:
            level, procpass = self.__getEvtLevel(
                tmpPath, [], level=0, start=False, end=False, processingpath="", startlevel=4
            )
        elif self.parameter_ == self.__bookkeepingParameters[3]:
            level, procpass = self.__getRunLevel(
                tmpPath, [], level=0, start=False, end=False, processingpath="", startlevel=1
            )
        else:
            level, procpass = self.__getLevel(
                path=tmpPath, visited=[], level=0, start=False, end=False, processingpath="", startlevel=3
            )
        self.treeLevels_ = level
        return level, processedPath, procpass

    #############################################################################
    # This method recursive visits all the tree nodes and found the processing pass
    def __getLevel(self, path, visited, level, start, end, processingpath, startlevel):
        """level."""
        for i in path:
            if level == startlevel and not start:
                for j in visited:
                    path.remove(j)
                level += 1
                return self.__getLevel(path, visited, level, start=True, end=False, processingpath="", startlevel=3)
            else:
                level += 1
                try:
                    int(i)
                except ValueError:
                    pass
                else:
                    if start:
                        end = True
                if start and not end:
                    level = startlevel
                    processingpath += "/" + i
                elif end and level <= startlevel + 1:
                    level = startlevel + 1
            visited += [i]
        return level, processingpath

    #############################################################################
    def __getRunLevel(self, path, visited, level, start, end, processingpath, startlevel):
        """run level."""
        for i in path:
            if level == startlevel and not start:
                for j in visited:
                    path.remove(j)
                level += 1
                return self.__getRunLevel(path, visited, level, start=True, end=False, processingpath="", startlevel=1)
            else:
                level += 1
                try:
                    int(i)
                except ValueError:
                    pass
                else:
                    if start:
                        end = True
                if start and not end:
                    level = startlevel
                    processingpath += "/" + i
                elif end and level <= startlevel + 1:
                    level = startlevel + 1
            visited += [i]
        return level, processingpath

    #############################################################################
    # This method recursive visite all the tree nodes and found the processing pass
    def __getEvtLevel(self, path, visited, level, start, end, processingpath, startlevel):
        """evt level."""
        for i in path:
            if level == startlevel and not start:
                for j in visited:
                    path.remove(j)
                level += 1
                return self.__getEvtLevel(path, visited, level, start=True, end=False, processingpath="", startlevel=4)
            else:
                level += 1
                try:
                    int(i)
                except ValueError:
                    if start and i in self.__filetypes:
                        end = True
                else:
                    if start:
                        end = True
                if start and not end:
                    level = startlevel
                    processingpath += "/" + i
                elif end and level <= startlevel + 1:
                    level = startlevel + 1
            visited += [i]
        return level, processingpath

    #############################################################################
    def _listConfigs(self, path, sortDict, startItem, maxitems):
        """list 1th tree."""
        entityList = list()
        levels, processedPath, procpass = self.getLevelAndPath(path)

        if levels == 0:  # configname
            self.clevelHeader_0()
            entityList += self.clevelBody_0(
                path,
                levels,
            )

        if levels == 1:  # config version
            in_dict = self.clevelHeader_1(processedPath)
            entityList += self.clevelBody_1(path, levels, in_dict)

        if levels == 2:  # sim or data desc
            in_dict = self.clevelHeader_2(processedPath)
            entityList += self.clevelBody_2(path, levels, in_dict)

        if levels == 3:  # processing
            in_dict = self.clevelHeader_3(processedPath)
            entityList += self.clevelBody_3(path, levels, in_dict, procpass)

        if levels == 4 and self.advancedQuery_:  # prod
            in_dict = self.clevelHeader_4(processedPath, procpass)
            entityList += self.clevelBody_4(path, levels, in_dict)
        elif levels == 4 and not self.advancedQuery_:
            processedPath += ["ALL"]
            in_dict = self.clevelHeader_5(processedPath, procpass)
            entityList += self.clevelBody_5(path, 5, in_dict)

        if levels == 5 and self.advancedQuery_:  # file type
            in_dict = self.clevelHeader_5(processedPath, procpass)
            entityList += self.clevelBody_5(path, levels, in_dict)
        elif levels == 5 and not self.advancedQuery_:
            levels = 6

        if levels == 6 and startItem == 0 and maxitems == 0:  # files
            in_dict = self.clevelHeader_6(processedPath, procpass)
            entityList += self.clevelBody_6(path, levels, in_dict)
        elif levels == 6 and (startItem != 0 or maxitems != 0):  # files
            in_dict = self.clevelHeader_6(processedPath, procpass)
            entityList += self.clevelBodyLimited_6(path, levels, in_dict, sortDict, startItem, maxitems)
        return entityList

    #############################################################################
    def __addAll(self, path, levels, description):
        """level all."""
        if self.advancedQuery_:
            return self._getEntityFromPath(path, "ALL", levels, description)

    @staticmethod
    def __createPath(processedPath, name):
        """create a path."""
        path = ""
        for i in processedPath:
            string = "/" + i[0] + "_" + i[1]
            path += string

        path += "/" + name[0] + "_" + name[1]
        return path

    #############################################################################
    @staticmethod
    def clevelHeader_0():
        """configuration tree (first tree)"""
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Configurations names:")
        gLogger.debug("-----------------------------------------------------------")

        # list root
        gLogger.debug("listing Configuration Names")

    #############################################################################
    def clevelBody_0(self, path, levels):
        """1t level of the configuration tree."""
        entityList = list()
        result = self.db_.getAvailableConfigNames()

        if result["OK"]:
            dbResult = result["Value"]
            for record in dbResult["Records"]:
                entityList += [self._getEntityFromPath(path, record[0], levels, None, {}, "getAvailableConfigNames")]
            self._cacheIt(entityList)
        else:
            gLogger.error(result["Message"])
        return entityList

    #############################################################################
    @staticmethod
    def clevelHeader_1(processedPath):
        """second level."""
        gLogger.debug("listing configversions")
        in_dict = {"ConfigName": processedPath[0]}
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Selected parameters:")
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug(f"Configuration Name      | {processedPath[0]} ")

        gLogger.debug("Available Config Versions:")
        return in_dict

    def clevelBody_1(self, path, levels, in_dict):
        """second."""
        entityList = list()
        result = self.db_.getConfigVersions(in_dict)
        if result["OK"]:
            dbResult = result["Value"]
            description = dbResult["ParameterNames"][0]
            for record in dbResult["Records"]:
                entityList += [
                    self._getEntityFromPath(path, record[0], levels, description, in_dict, "getConfigVersions")
                ]
            self._cacheIt(entityList)
        else:
            gLogger.error(result["Message"])
        return entityList

    #############################################################################
    @staticmethod
    def clevelHeader_2(processedPath):
        """third."""
        gLogger.debug("listing Simulation Conditions!")
        in_dict = {"ConfigName": processedPath[0], "ConfigVersion": processedPath[1]}

        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Selected parameters:")
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug(in_dict)
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Available Simulation Conditions:")
        return in_dict

    #############################################################################
    def clevelBody_2(self, path, levels, in_dict):
        """build the tree node."""
        entityList = list()
        result = self.db_.getConditions(in_dict)
        if result["OK"]:
            dbResult = result["Value"]
            if dbResult[0]["TotalRecords"] > 0:
                add = self.__addAll(path, levels, "Simulation Conditions/DataTaking")
                if add:
                    entityList += [add]
                for record in dbResult[0]["Records"]:
                    value = {}
                    j = 0
                    for i in dbResult[0]["ParameterNames"]:
                        value[i] = record[j]
                        j += 1
                    entityList += [
                        self._getSpecificEntityFromPath(
                            path,
                            value,
                            record[1],
                            levels,
                            None,
                            "Simulation Conditions/DataTaking",
                            in_dict,
                            "getConditions",
                        )
                    ]
                self._cacheIt(entityList)
            if dbResult[1]["TotalRecords"] > 0:
                for record in dbResult[1]["Records"]:
                    value = {}
                    j = 0
                    for i in dbResult[1]["ParameterNames"]:
                        value[i] = record[j]
                        j += 1
                    entityList += [
                        self._getSpecificEntityFromPath(
                            path,
                            value,
                            record[1],
                            levels,
                            None,
                            "Simulation Conditions/DataTaking",
                            in_dict,
                            "getConditions",
                        )
                    ]
                self._cacheIt(entityList)

        else:
            gLogger.error(result["Message"])

        return entityList

    #############################################################################
    @staticmethod
    def clevelHeader_3(processedPath):
        """fourth level."""
        gLogger.debug("listing processing pass")
        in_dict = {
            "ConfigName": processedPath[0],
            "ConfigVersion": processedPath[1],
            "ConditionDescription": processedPath[2],
        }

        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Selected parameters:")
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug(in_dict)
        gLogger.debug("-----------------------------------------------------------")

        gLogger.debug("Available processing pass:\n")
        return in_dict

    #############################################################################
    def clevelBody_3(self, path, levels, in_dict, procpass):
        """fourth tree node."""
        entityList = list()
        in_dict["ProcessingPass"] = procpass
        result = self.db_.getProcessingPass(in_dict, procpass)
        if result["OK"]:
            dbResult = result["Value"]
            if dbResult[0]["TotalRecords"] > 0:  # it is a processing pass
                add = self.__addAll(path, levels, "Processing Pass")
                if add:
                    entityList += [add]
                for record in dbResult[0]["Records"]:
                    entityList += [
                        self._getEntityFromPath(
                            path, record[0], levels, "Processing Pass", in_dict, "getProcessingPass"
                        )
                    ]
                self._cacheIt(entityList)
            if dbResult[1]["TotalRecords"] > 0:
                value = {}
                for record in dbResult[1]["Records"]:
                    value = {"Event Type": record[0], "Description": record[1]}
                    entityList += [
                        self._getSpecificEntityFromPath(
                            path, value, str(record[0]), levels, None, "Event types", in_dict, "getProcessingPass"
                        )
                    ]
                self._cacheIt(entityList)
        else:
            gLogger.error(result["Message"])
        return entityList

    #############################################################################
    @staticmethod
    def clevelHeader_4(processedPath, procpass):
        """5th level."""
        gLogger.debug("listing event types")
        retVal = procpass.split("/")[1:]
        for i in retVal:
            processedPath.remove(i)

        in_dict = {
            "ConfigName": processedPath[0],
            "ConfigVersion": processedPath[1],
            "ConditionDescription": processedPath[2],
            "EventType": processedPath[3],
        }
        in_dict["ProcessingPass"] = procpass

        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Selected parameters: ")
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug(in_dict)
        gLogger.debug("-----------------------------------------------------------")

        gLogger.debug("Available event types types:")
        return in_dict

    #############################################################################
    def clevelBody_4(self, path, levels, in_dict):
        """5th tree node."""
        entityList = list()
        result = self.db_.getProductions(in_dict)
        if result["OK"]:
            dbResult = result["Value"]
            for record in dbResult["Records"]:
                entityList += [
                    self._getEntityFromPath(
                        path, str(record[0]), levels, "Production(s)/Run(s)", in_dict, "getProductions"
                    )
                ]
            self._cacheIt(entityList)
        else:
            gLogger.error(result["Message"])
        return entityList

    #############################################################################
    def clevelHeader_5(self, processedPath, procpass):
        """6th tree."""
        gLogger.debug("listing event types")
        retVal = procpass.split("/")[1:]
        for i in retVal:
            processedPath.remove(i)

        if self.advancedQuery_:
            in_dict = {
                "ConfigName": processedPath[0],
                "ConfigVersion": processedPath[1],
                "ConditionDescription": processedPath[2],
                "EventType": processedPath[3],
                "Production": processedPath[4],
            }
        else:
            in_dict = {
                "ConfigName": processedPath[0],
                "ConfigVersion": processedPath[1],
                "ConditionDescription": processedPath[2],
                "EventType": processedPath[3],
            }
        in_dict["ProcessingPass"] = procpass

        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Selected parameters: ")
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug(in_dict)
        gLogger.debug("-----------------------------------------------------------")

        gLogger.debug("Available event types types:")
        return in_dict

    #############################################################################
    def clevelBody_5(self, path, levels, in_dict):
        """build the 6th tree node."""
        entityList = list()
        in_dict["Visible"] = "Y"
        result = self.db_.getFileTypes(in_dict)
        if result["OK"]:
            dbResult = result["Value"]
            for record in dbResult["Records"]:
                entityList += [self._getEntityFromPath(path, record[0], levels, "FileTypes", in_dict, "getFileTypes")]
            self._cacheIt(entityList)
        else:
            gLogger.error(result["Message"])
        return entityList

    #############################################################################
    def clevelHeader_6(self, processedPath, procpass):
        """7th tree prepare."""
        gLogger.debug("listing event types")

        retVal = procpass.split("/")[1:]
        for i in retVal:
            processedPath.remove(i)

        if self.advancedQuery_:
            in_dict = {
                "ConfigName": processedPath[0],
                "ConfigVersion": processedPath[1],
                "ConditionDescription": processedPath[2],
                "EventType": processedPath[3],
                "Production": processedPath[4],
                "FileType": processedPath[5],
            }
        else:
            in_dict = {
                "ConfigName": processedPath[0],
                "ConfigVersion": processedPath[1],
                "ConditionDescription": processedPath[2],
                "EventType": processedPath[3],
                "FileType": processedPath[4],
            }
        in_dict["ProcessingPass"] = procpass

        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Selected parameters: ")
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug(in_dict)
        gLogger.debug("-----------------------------------------------------------")

        return in_dict

    #############################################################################
    def clevelBody_6(self, path, levels, in_dict):
        """build the 7th tree."""
        entityList = list()
        in_dict["DataQuality"] = self.__getSelectedQualities()
        in_dict["Visible"] = "Y"
        result = self.db_.getFilesWithMetadata(in_dict)
        if result["OK"]:
            for record in result["Value"]["Records"]:
                value = {
                    "name": record[0],
                    "EventStat": record[1],
                    "FileSize": record[2],
                    "CreationDate": record[3],
                    "JobStart": record[4],
                    "JobEnd": record[5],
                    "WorkerNode": record[6],
                    "FileType": in_dict["FileType"],
                    "RunNumber": record[8],
                    "FillNumber": record[9],
                    "FullStat": record[10],
                    "DataqualityFlag": record[11],
                    "EventType": in_dict["EventType"],
                    "EventInputStat": record[12],
                    "TotalLuminosity": record[13],
                    "Luminosity": record[14],
                    "InstLuminosity": record[15],
                    "TCK": record[16],
                }
                self.files_ += [record[0]]
                entityList += [
                    self._getEntityFromPath(path, value, levels, "List of files", in_dict, "getFilesWithMetadata")
                ]
            self._cacheIt(entityList)
        else:
            return result
        return entityList

    #############################################################################
    def clevelBodyLimited_6(self, path, levels, in_dict, sortDict, startItem, maxitems):
        """7th tree node for web."""
        entityList = list()
        in_dict["DataQuality"] = self.__getSelectedQualities()
        result = self.__getFiles(in_dict, sortDict, startItem, maxitems)
        for record in result["Records"]:
            value = {
                "name": record[0],
                "EventStat": record[1],
                "FileSize": record[2],
                "CreationDate": record[3],
                "JobStart": record[4],
                "JobEnd": record[5],
                "WorkerNode": record[6],
                "FileType": in_dict["FileType"],
                "EventType": in_dict["EventType"],
                "RunNumber": record[9],
                "FillNumber": record[10],
                "FullStat": record[11],
                "DataqualityFlag": record[12],
                "EventInputStat": record[13],
                "TotalLuminosity": record[14],
                "Luminosity": record[15],
                "InstLuminosity": record[16],
                "TCK": record[17],
            }

            self.files_ += [record[0]]
            entityList += [
                self._getEntityFromPath(path, value, levels, "List of files", in_dict, "getFilesWithMetadata")
            ]
        self._cacheIt(entityList)
        return entityList

    #############################################################################
    def _listEventTypes(self, path, sortDict, startItem, maxitems):
        """second tree based on event type."""
        entityList = list()
        levels, processedPath, procpass = self.getLevelAndPath(path)

        if levels == 0:  # configname
            self.clevelHeader_0()
            entityList += self.clevelBody_0(
                path,
                levels,
            )

        if levels == 1:  # config version
            in_dict = self.clevelHeader_1(processedPath)
            entityList += self.clevelBody_1(path, levels, in_dict)

        if levels == 2:  # event type
            in_dict = self.clevelHeader_2(processedPath)
            entityList += self.elevelBody_2(path, levels, in_dict)

        if levels == 3:  # sim ot daq desq
            in_dict = self.elevelHeader_3(processedPath)
            entityList += self.elevelBody_3(path, levels, in_dict)

        if levels == 4:  # processing pass
            in_dict = self.elevelHeader_4(processedPath)
            entityList += self.elevelBody_4(path, levels, in_dict, procpass)

        if self.advancedQuery_ and levels == 5:
            in_dict = self.elevelHedaer_5(processedPath, procpass)
            entityList += self.clevelBody_5(path, levels, in_dict)
        elif levels == 5:
            levels = 6

        if levels == 6 and startItem == 0 and maxitems == 0:  # files
            in_dict = self.elevelHeader_6(processedPath, procpass)
            entityList += self.clevelBody_6(path, levels, in_dict)
        elif levels == 6 and (startItem != 0 or maxitems != 0):  # files
            in_dict = self.elevelHeader_6(processedPath, procpass)
            entityList += self.clevelBodyLimited_6(path, levels, in_dict, sortDict, startItem, maxitems)

        return entityList

    #############################################################################
    def elevelBody_2(self, path, levels, in_dict):
        """event type based queries."""
        entityList = list()
        result = self.db_.getEventTypes(in_dict)
        if result["OK"]:
            dbResult = result["Value"]
            if len(dbResult) > 1:
                add = self.__addAll(path, levels, "Event types")
                if add:
                    entityList += [add]
            for record in dbResult["Records"]:
                value = {"Event Type": record[0], "Description": record[1]}
                entityList += [
                    self._getSpecificEntityFromPath(
                        path, value, str(record[0]), levels, None, "Event types", in_dict, "getEventTypes"
                    )
                ]
            self._cacheIt(entityList)
        else:
            gLogger.error(result["Message"])
            return result
        return entityList

    #############################################################################
    @staticmethod
    def elevelHeader_3(processedPath):
        """event type based tree node."""
        gLogger.debug("listing simulation conditions")

        in_dict = {"ConfigName": processedPath[0], "ConfigVersion": processedPath[1], "EventType": processedPath[2]}

        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Selected parameters:")
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug(in_dict)
        gLogger.debug("-----------------------------------------------------------")

        gLogger.debug("Available conditions:")
        return in_dict

    #############################################################################
    def elevelBody_3(self, path, levels, in_dict):
        """building the tree node."""

        entityList = list()

        result = self.db_.getConditions(in_dict)
        if result["OK"]:
            dbResult = result["Value"]
            if dbResult[0]["TotalRecords"] > 0:
                add = self.__addAll(path, levels, "Simulation Conditions/DataTaking")
                if add:
                    entityList += [add]
                for record in dbResult[0]["Records"]:
                    value = {}
                    j = 0
                    for i in dbResult[0]["ParameterNames"]:
                        value[i] = record[j]
                        j += 1
                    entityList += [
                        self._getSpecificEntityFromPath(
                            path,
                            value,
                            record[1],
                            levels,
                            None,
                            "Simulation Conditions/DataTaking",
                            in_dict,
                            "getConditions",
                        )
                    ]
                self._cacheIt(entityList)
            if dbResult[1]["TotalRecords"] > 0:
                for record in dbResult[1]["Records"]:
                    value = {}
                    j = 0
                    for i in dbResult[1]["ParameterNames"]:
                        value[i] = record[j]
                        j += 1
                    entityList += [
                        self._getSpecificEntityFromPath(
                            path,
                            value,
                            record[1],
                            levels,
                            None,
                            "Simulation Conditions/DataTaking",
                            in_dict,
                            "getConditions",
                        )
                    ]
                self._cacheIt(entityList)

        else:
            gLogger.error(result["Message"])

        return entityList

    #############################################################################
    @staticmethod
    def elevelHeader_4(processedPath):
        """prepare the 5th query."""
        gLogger.debug("listing processing pass")

        in_dict = {
            "ConfigName": processedPath[0],
            "ConfigVersion": processedPath[1],
            "EventType": processedPath[2],
            "ConditionDescription": processedPath[3],
        }

        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Selected parameters: ")
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug(in_dict)
        gLogger.debug("-----------------------------------------------------------")

        gLogger.debug("Available processing pass types:")
        return in_dict

    #############################################################################
    def elevelBody_4(self, path, levels, in_dict, procpass):
        """make the tree node."""
        entityList = list()

        result = self.db_.getProcessingPass(in_dict, procpass)
        if result["OK"]:
            dbResult = result["Value"]
            if dbResult[0]["TotalRecords"] > 0:  # it is a processing pass
                add = self.__addAll(path, levels, "Processing Pass")
                if add:
                    entityList += [add]
                for record in dbResult[0]["Records"]:
                    entityList += [
                        self._getEntityFromPath(
                            path, record[0], levels, "Processing Pass", in_dict, "getProcessingPass"
                        )
                    ]
                self._cacheIt(entityList)
            elif self.advancedQuery_:
                in_dict["ProcessingPass"] = procpass
                result = self.db_.getProductions(in_dict)
                if result["OK"]:
                    dbResult = result["Value"]
                    for record in dbResult["Records"]:
                        entityList += [
                            self._getEntityFromPath(
                                path, str(record[0]), levels, "Production(s)/Run(s)", in_dict, "getProductions"
                            )
                        ]
                    self._cacheIt(entityList)

        if len(procpass) > 0:
            in_dict["ProcessingPass"] = procpass
            in_dict["Visible"] = "Y"
            result = self.db_.getFileTypes(in_dict)
            if result["OK"]:
                dbResult = result["Value"]
                for record in dbResult["Records"]:
                    entityList += [
                        self._getEntityFromPath(path, record[0], levels, "FileTypes", in_dict, "getFileTypes")
                    ]
                self._cacheIt(entityList)
            else:
                gLogger.error(result["Message"])
        return entityList

    #############################################################################
    def elevelHedaer_5(self, processedPath, procpass):
        """prepare tree node."""
        retVal = procpass.split("/")[1:]
        for i in retVal:
            processedPath.remove(i)

        if self.advancedQuery_:
            in_dict = {
                "ConfigName": processedPath[0],
                "ConfigVersion": processedPath[1],
                "EventType": processedPath[2],
                "ConditionDescription": processedPath[3],
                "Production": processedPath[4],
            }
        else:
            in_dict = {
                "ConfigName": processedPath[0],
                "ConfigVersion": processedPath[1],
                "EventType": processedPath[2],
                "ConditionDescription": processedPath[3],
            }
        in_dict["ProcessingPass"] = procpass

        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Selected parameters:   ")
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug(in_dict)
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Available Production(s):")
        return in_dict

    #############################################################################
    def elevelHeader_6(self, processedPath, procpass):
        """prepare tree mode."""
        retVal = procpass.split("/")[1:]
        for i in retVal:
            processedPath.remove(i)

        if self.advancedQuery_:
            in_dict = {
                "ConfigName": processedPath[0],
                "ConfigVersion": processedPath[1],
                "EventType": processedPath[2],
                "ConditionDescription": processedPath[3],
                "Production": processedPath[4],
                "FileType": processedPath[5],
            }
        else:
            in_dict = {
                "ConfigName": processedPath[0],
                "ConfigVersion": processedPath[1],
                "EventType": processedPath[2],
                "ConditionDescription": processedPath[3],
                "FileType": processedPath[4],
            }
        in_dict["ProcessingPass"] = procpass

        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Selected parameters:   ")
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug(in_dict)
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Available file types:")
        return in_dict

    #############################################################################
    def _listProduction(self, path):
        """production lookup."""
        entityList = list()
        levels, processedPath, procpass = self.getLevelAndPath(path)
        gLogger.debug(str(procpass))

        if levels == 0:
            self.plevelHeader_0()
            entityList += self.plevelBody_0(path, levels)

        if levels == 1:
            in_dict = self.plevelHeader_2(processedPath)
            entityList += self.plevelBody_2(path, levels, in_dict)

        if levels == 2:
            in_dict = self.plevelHeader_3(processedPath)
            entityList += self.plevelBody_3(path, levels, in_dict)

        if levels == 3:
            in_dict = self.plevelHeader_4(processedPath)
            entityList += self.plevelBody_4(path, levels, in_dict)

        return entityList

    #############################################################################
    def _listRuns(self, path):
        """run lookup."""
        entityList = list()

        levels, processedPath, procpass = self.getLevelAndPath(path)

        if levels == 0:
            self.rlevelHeader_0()
            entityList += self.rlevelBody_0(path, levels)

        if levels == 1:
            in_dict = self.rlevelHeader_2(processedPath)
            entityList += self.rlevelBody_2(path, levels, in_dict, procpass)

        if levels == 2:
            in_dict = self.rlevelHeader_3(processedPath, procpass)
            entityList += self.rlevelBody_3(path, levels, in_dict)

        if levels == 3:
            in_dict = self.rlevelHeader_4(processedPath, procpass)
            entityList += self.clevelBody_6(path, levels, in_dict)

        return entityList

    #############################################################################
    @staticmethod
    def plevelHeader_0():
        """prepare production lookup tree node."""
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("productions:")
        gLogger.debug("-----------------------------------------------------------")

        # list root
        gLogger.debug("listing productions")

    #############################################################################
    def plevelBody_0(self, path, levels):
        """make the node of the production lookup tree."""
        entityList = list()
        result = self.db_.getAvailableProductions()

        if result["OK"]:
            dbResult = result["Value"]
            for record in dbResult:
                prod = str(record[0])
                entityList += [self._getEntityFromPath(path, str(prod), levels, "Production(s)/Run(s)")]
            self._cacheIt(entityList)
        else:
            gLogger.error(result["Message"])
        return entityList

    #############################################################################
    @staticmethod
    def plevelHeader_2(processedPath):
        """prepare the tree node."""
        gLogger.debug("listing eventtype")

        in_dict = {"Production": processedPath[0]}

        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Selected parameters:")
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug(in_dict)

        gLogger.debug("Available event types:")
        return in_dict

    #############################################################################
    def plevelBody_2(self, path, levels, in_dict):
        """make the tree node."""
        entityList = list()

        result = self.db_.getEventTypes(in_dict)
        if result["OK"]:
            dbResult = result["Value"]
            if len(dbResult) > 1:
                add = self.__addAll(path, levels, "Event types")
                if add:
                    entityList += [add]
            for record in dbResult["Records"]:
                value = {"Event Type": record[0], "Description": record[1]}
                entityList += [
                    self._getSpecificEntityFromPath(
                        path, value, str(record[0]), levels, None, "Event types", in_dict, "getEventTypes"
                    )
                ]
            self._cacheIt(entityList)
        else:
            gLogger.error(result["Message"])
        return entityList

    #############################################################################
    @staticmethod
    def plevelHeader_3(processedPath):
        """prepare tree node."""
        gLogger.debug("listing file types")
        in_dict = {"Production": processedPath[0], "EventType": processedPath[1]}
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Selected parameters:")
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug(in_dict)

        gLogger.debug("Available file types:")
        return in_dict

    #############################################################################
    def plevelBody_3(self, path, levels, in_dict):
        """make tree node."""
        entityList = list()
        in_dict["Visible"] = "Y"
        result = self.db_.getFileTypes(in_dict)
        if result["OK"]:
            dbResult = result["Value"]
            for record in dbResult["Records"]:
                entityList += [self._getEntityFromPath(path, record[0], levels, "FileTypes", in_dict, "getFileTypes")]
            self._cacheIt(entityList)
        else:
            gLogger.error(result["Message"])
        return entityList

    #############################################################################
    @staticmethod
    def plevelHeader_4(processedPath):
        """prepare the tree node."""
        gLogger.debug("listing file types")
        in_dict = {"Production": processedPath[0], "EventType": processedPath[1], "FileType": processedPath[2]}

        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Selected parameters:")
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug(in_dict)

        gLogger.debug("Available files:")
        return in_dict

    #############################################################################
    def plevelBody_4(self, path, levels, in_dict):
        """make the tree node."""
        entityList = list()
        in_dict["DataQuality"] = self.__getSelectedQualities()
        in_dict["Visible"] = "Y"
        result = self.db_.getFilesWithMetadata(in_dict)
        if result["OK"]:
            for record in result["Value"]["Records"]:
                value = {
                    "name": record[0],
                    "EventStat": record[1],
                    "FileSize": record[2],
                    "CreationDate": record[3],
                    "JobStart": record[4],
                    "JobEnd": record[5],
                    "WorkerNode": record[6],
                    "FileType": in_dict["FileType"],
                    "RunNumber": record[8],
                    "FillNumber": record[9],
                    "FullStat": record[10],
                    "DataqualityFlag": record[11],
                    "EventType": in_dict["EventType"],
                    "EventInputStat": record[12],
                    "TotalLuminosity": record[13],
                    "Luminosity": record[14],
                    "InstLuminosity": record[15],
                    "TCK": record[16],
                }
                self.files_ += [record[0]]
                entityList += [
                    self._getEntityFromPath(path, value, levels, "List of files", in_dict, "getFilesWithMetadata")
                ]
            self._cacheIt(entityList)
        else:
            return result

        return entityList

    @staticmethod
    def rlevelHeader_0():
        """prepare run lookup node."""
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Runs:")
        gLogger.debug("-----------------------------------------------------------")

        # list root
        gLogger.debug("listing runs")

    #############################################################################
    def rlevelBody_0(self, path, levels):
        """make tree node."""
        entityList = list()
        result = self.db_.getAvailableRuns()

        if result["OK"]:
            dbResult = result["Value"]
            for record in dbResult:
                run = str(record[0])
                entityList += [self._getEntityFromPath(path, str(run), levels, "Run(s)")]
            self._cacheIt(entityList)
        else:
            gLogger.error(result["Message"])
        return entityList

    #############################################################################
    @staticmethod
    def rlevelHeader_2(processedPath):
        """prepare tree node."""
        gLogger.debug("listing processing pass")
        in_dict = {"RunNumber": processedPath[0]}
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Selected parameters:")
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug(in_dict)

        gLogger.debug("Available processing pass:")
        return in_dict

    #############################################################################
    def rlevelBody_2(self, path, levels, in_dict, procpass):
        """make tree node."""
        entityList = list()
        in_dict["ProcessingPass"] = procpass
        result = self.db_.getProcessingPass(in_dict, procpass)
        if result["OK"]:
            dbResult = result["Value"]
            if dbResult[0]["TotalRecords"] > 0:  # it is a processing pass
                add = self.__addAll(path, levels, "Processing Pass")
                if add:
                    entityList += [add]
            for record in dbResult[0]["Records"]:
                entityList += [
                    self._getEntityFromPath(path, record[0], levels, "Processing Pass", in_dict, "getProcessingPass")
                ]
            self._cacheIt(entityList)
            if dbResult[1]["TotalRecords"] > 0:
                value = {}
                for record in dbResult[1]["Records"]:
                    value = {"Event Type": record[0], "Description": record[1]}
                    entityList += [
                        self._getSpecificEntityFromPath(
                            path, value, str(record[0]), levels, None, "Event types", in_dict, "getProcessingPass"
                        )
                    ]
                self._cacheIt(entityList)
        else:
            gLogger.error(result["Message"])
        return entityList

    #############################################################################
    @staticmethod
    def rlevelHeader_3(processedPath, procpass):
        """prepare node of the run lookup tree."""
        gLogger.debug("listing eventtypes")
        retVal = procpass.split("/")[1:]
        for i in retVal:
            processedPath.remove(i)

        in_dict = {"RunNumber": processedPath[0], "EventType": processedPath[1]}
        in_dict["ProcessingPass"] = procpass

        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Selected parameters:")
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug(in_dict)

        gLogger.debug("Available eventtypes types:")
        return in_dict

    #############################################################################
    def rlevelBody_3(self, path, levels, in_dict):
        """make tree node."""
        entityList = list()
        in_dict["Visible"] = "Y"
        result = self.db_.getFileTypes(in_dict)
        if result["OK"]:
            dbResult = result["Value"]
            for record in dbResult["Records"]:
                entityList += [self._getEntityFromPath(path, record[0], levels, "FileTypes", in_dict, "getFileTypes")]
            self._cacheIt(entityList)
        else:
            gLogger.error(result["Message"])
        return entityList

    #############################################################################
    @staticmethod
    def rlevelHeader_4(processedPath, procpass):
        """prepare tree."""
        gLogger.debug("listing file types")
        retVal = procpass.split("/")[1:]
        for i in retVal:
            processedPath.remove(i)

        in_dict = {"RunNumber": processedPath[0], "EventType": processedPath[1], "FileType": processedPath[2]}
        in_dict["ProcessingPass"] = procpass
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug("Selected parameters:")
        gLogger.debug("-----------------------------------------------------------")
        gLogger.debug(in_dict)

        gLogger.debug("Available file types:")
        return in_dict

    #############################################################################
    @staticmethod
    def _getEntityFromPath(presentPath, newPathElement, level, leveldescription=None, selection=None, method=None):
        """create a entity."""
        if isinstance(newPathElement, dict):
            # this must be a file
            entity = objects.Entity(newPathElement)
            newPathElement = str(entity["name"]).rsplit("/", 1)[1]
            entity.update({"FileName": entity["name"]})
            expandable = False
            entity.update({"expandable": expandable})
            if selection is not None:
                entity.update({"selection": selection})

            if method is not None:
                entity.update({"method": method})

        else:
            # this must be a folder
            entity = objects.Entity()
            name = newPathElement

            expandable = True

            fullPath = presentPath.rstrip(INTERNAL_PATH_SEPARATOR)
            fullPath += INTERNAL_PATH_SEPARATOR + name

            entity.update({"name": name, "fullpath": fullPath, "expandable": expandable})

            if leveldescription is not None:
                entity.update({"level": leveldescription})

            if leveldescription == "FileTypes":
                entity.update({"showFiles": 0})

            if selection is not None:
                entity.update({"selection": selection})

            if method is not None:
                entity.update({"method": method})

            elif level == 5:
                entity.update({"showFiles": 0})
        return entity

    #############################################################################
    @staticmethod
    def _getSpecificEntityFromPath(
        presentPath, value, newPathElement, level, description=None, leveldescription=None, selection=None, method=None
    ):
        """crate a specific entity."""
        if isinstance(value, dict):
            entity = objects.Entity(value)
            name = newPathElement

            expandable = True

            fullPath = presentPath.rstrip(INTERNAL_PATH_SEPARATOR)
            fullPath += INTERNAL_PATH_SEPARATOR + name

            if description is not None:
                entity.update({"name": description, "fullpath": fullPath, "expandable": expandable})
            else:
                entity.update({"name": name, "fullpath": fullPath, "expandable": expandable})
            if leveldescription is not None:
                entity.update({"level": leveldescription})

            if selection is not None:
                entity.update({"selection": selection})

            if method is not None:
                entity.update({"method": method})

            if level == 5:
                entity.update({"showFiles": 0})

        return entity

    #############################################################################
    def _processPath(self, path):
        """takes an absolute path and returns of tuples with prefixes and posfixes
        of path elements.

        If invalid path returns null
        """
        path = path.strip(INTERNAL_PATH_SEPARATOR + " ")
        return path.split(INTERNAL_PATH_SEPARATOR)

    #############################################################################
    def _cacheIt(self, entityList):
        """it caches an entity or a list of entities."""
        if isinstance(entityList, objects.Entity):
            # convert it into a list
            entityList = [entityList]
        elif not isinstance(entityList, list):
            # neither entity nor list
            gLogger.warn("couldn't cache invalid entity(list) of type " + str(entityList.__class__))
            return

        for entity in entityList:
            # TO IMPLEMENT!! time of the caching
            try:
                if "fullpath" in entity:
                    self.__entityCache.update({entity["fullpath"]: (entity, 0)})
            except ValueError as ex:
                gLogger.warn("couldn't cache entity(?) " + str(entity) + "  " + str(ex))
                return S_ERROR("couldnt cache entity!")

    #############################################################################
    @staticmethod
    def getAbsolutePath(path):
        """get current working directory if empty."""
        if path in ["", ".", None]:
            path = INTERNAL_PATH_SEPARATOR  # root
            # convert it into absolute path
        path = os.path.normpath(path)
        if os.sep != INTERNAL_PATH_SEPARATOR:
            path = path.replace(os.sep, INTERNAL_PATH_SEPARATOR)

        # for this special case of anomaly when double // may appear
        path = path.replace(2 * INTERNAL_PATH_SEPARATOR, INTERNAL_PATH_SEPARATOR)

        return S_OK(path)

    #############################################################################
    def get(self, path="/"):
        """get a node for a given path."""
        path = self.getAbsolutePath(path)["Value"]
        entity = self._getEntity(path)
        if not isinstance(entity, objects.Entity):
            gLogger.error(path + " doesn't exist!")
            # raise ValueError, "Invalid path %s" % path
        return S_OK(entity)

    #############################################################################
    def _getEntity(self, path):
        """This is not doing anything at the moment."""
        try:
            entity = self.__entityCache[path][0]
            gLogger.debug("getting " + str(path) + " from the cache")
            return entity
        except ValueError as ex:
            # not cached so far
            gLogger.debug(str(path) + " not in cache. Fetching..." + ex)

        # Second try

        try:
            gLogger.debug("getting " + str(path) + " eventually from the cache")
            entity = self.__entityCache[path][0]
            return entity
        except ValueError as ex:
            # still not in the cache... wrong path
            gLogger.warn(str(path) + " seems to be a wrong path" + ex)
            return None

        return entity

    #############################################################################
    @staticmethod
    def getNumberOfEvents(files):
        """statistics."""
        esum = 0
        for lfn in files:
            esum += int(lfn["EventStat"])
        return esum

    #############################################################################
    def getJobInfo(self, lfn):
        """job info."""
        result = self.db_.getJobInfo(lfn)
        value = None
        if result["OK"]:
            dbResult = result["Value"]
            for record in dbResult:
                value = {
                    "DiracJobID": record[0],
                    "DiracVersion": record[1],
                    "EventInputStat": record[2],
                    "ExecTime": record[3],
                    "FirstEventNumber": record[4],
                    "Location": record[5],
                    "Name": record[6],
                    "NumberofEvents": record[7],
                    "StatisticsRequested": record[8],
                    "WNCPUPOWER": record[9],
                    "CPUTime": record[10],
                    "WNCACHE": record[11],
                    "WNMEMORY": record[12],
                    "WNMODEL": record[13],
                    "WORKERNODE": record[14],
                    "WNCPUHS06": record[15],
                    "TotalLuminosity": record[17],
                }
        else:
            gLogger.error(result["Message"])
        return value

    #############################################################################
    def getLimitedFiles(self, selectionDict, sortDict, startItem, maxitems):
        """web."""
        if self.parameter_ == self.__bookkeepingParameters[0]:
            return self._getLimitedFilesConfigParams(selectionDict, sortDict, startItem, maxitems)
        elif self.parameter_ == self.__bookkeepingParameters[1]:
            return self._getLimitedFilesEventTypeParams(selectionDict, sortDict, startItem, maxitems)
        elif self.parameter_ == self.__bookkeepingParameters[2]:
            return self._getLimitedFilesProductions(selectionDict, sortDict, startItem, maxitems)
        elif self.parameter_ == self.__bookkeepingParameters[3]:
            return self._getLimitedFilesRuns(selectionDict, sortDict, startItem, maxitems)

    #############################################################################
    def _getDataSetTree1(self, selectionDict):
        # it is the configname configversion(default) query.
        # The input parameter is a path and it constructs the dictionary.
        """input dictionary."""
        path = selectionDict["fullpath"]
        levels, processedPath, procpass = self.getLevelAndPath(path)
        gLogger.debug(str(levels))
        retVal = procpass.split("/")[1:]
        for i in retVal:
            processedPath.remove(i)

        if self.advancedQuery_:
            in_dict = {
                "ConfigName": processedPath[0],
                "ConfigVersion": processedPath[1],
                "ConditionDescription": processedPath[2],
                "EventType": processedPath[3],
                "Production": processedPath[4],
                "FileType": processedPath[5],
            }
        else:
            in_dict = {
                "ConfigName": processedPath[0],
                "ConfigVersion": processedPath[1],
                "ConditionDescription": processedPath[2],
                "EventType": processedPath[3],
                "FileType": processedPath[4],
            }
        in_dict["ProcessingPass"] = procpass
        in_dict["fullpath"] = path
        return in_dict

    #############################################################################
    def _getLimitedFilesConfigParams(self, selectionDict, sortDict, startItem, maxitems):
        """input dictionary."""
        selection = self._getDataSetTree1(selectionDict)
        return self.__getFiles(selection, sortDict, startItem, maxitems)

    #############################################################################
    def _getDataSetTree2(self, selectionDict):  # it is the event type based query
        """input dictionary of the event type tree."""
        path = selectionDict["fullpath"]
        levels, processedPath, procpass = self.getLevelAndPath(path)
        gLogger.debug(str(levels))
        retVal = procpass.split("/")[1:]
        for i in retVal:
            processedPath.remove(i)

        if self.advancedQuery_:
            in_dict = {
                "ConfigName": processedPath[0],
                "ConfigVersion": processedPath[1],
                "EventType": processedPath[2],
                "ConditionDescription": processedPath[3],
                "Production": processedPath[4],
                "FileType": processedPath[5],
            }
        else:
            in_dict = {
                "ConfigName": processedPath[0],
                "ConfigVersion": processedPath[1],
                "EventType": processedPath[2],
                "ConditionDescription": processedPath[3],
                "FileType": processedPath[4],
            }
        in_dict["ProcessingPass"] = procpass
        in_dict["fullpath"] = path
        return in_dict

    #############################################################################
    def _getLimitedFilesEventTypeParams(self, selectionDict, sortDict, startItem, maxitems):
        """input dictionary."""
        selection = self._getDataSetTree2(selectionDict)
        return self.__getFiles(selection, sortDict, startItem, maxitems)

    #############################################################################
    def _getDataSetTree3(self, selectionDict):  # production based query
        """input dictionary."""
        path = selectionDict["fullpath"]
        levels, processedPath, procpass = self.getLevelAndPath(path)
        gLogger.debug(str(levels))
        gLogger.debug(procpass)

        in_dict = {"Production": processedPath[0], "EventType": processedPath[1], "FileType": processedPath[2]}
        in_dict["fullpath"] = path
        return in_dict

    #############################################################################
    def _getLimitedFilesProductions(self, selectionDict, sortDict, startItem, maxitems):
        """input dictionary."""
        selection = self._getDataSetTree3(selectionDict)
        return self.__getFiles(selection, sortDict, startItem, maxitems)

    #############################################################################
    def _getDataSetTree4(self, selectionDict):  # run based query
        """input dictionary."""
        path = selectionDict["fullpath"]
        levels, processedPath, procpass = self.getLevelAndPath(path)
        gLogger.debug(str(levels))
        retVal = procpass.split("/")[1:]
        for i in retVal:
            processedPath.remove(i)

        in_dict = {
            "RunNumber": processedPath[0],
            "EventType": processedPath[1],
            "FileType": processedPath[2],
            "ProcessingPass": procpass,
        }
        in_dict["fullpath"] = path
        return in_dict

    #############################################################################
    def _getLimitedFilesRuns(self, selectionDict, sortDict, startItem, maxitems):
        """input dictionary."""
        selection = self._getDataSetTree4(selectionDict)
        return self.__getFiles(selection, sortDict, startItem, maxitems)

    #############################################################################
    def __getFiles(self, in_dict, sortDict, startItem, maxitems):
        """returns the files."""
        totalrecords = 0
        nbOfEvents = 0
        filesSize = 0
        lumi = 0
        selection = in_dict
        in_dict["Visible"] = "Y"
        if len(sortDict) > 0:
            in_dict["DataQuality"] = self.__getSelectedQualities()
            res = self.db_.getFilesSummary(in_dict)
            if not res["OK"]:
                gLogger.error(res["Message"])
            else:
                records = res["Value"]["Records"]
                totalrecords = records[0][0]
                nbOfEvents = records[0][1]
                filesSize = records[0][2]
                lumi = records[0][3]
        records = []
        parametersNames = []
        if startItem > -1 and maxitems != 0:
            in_dict["StartItem"] = startItem
            in_dict["MaxItem"] = maxitems
            in_dict["DataQuality"] = self.__getSelectedQualities()
            result = self.db_.getLimitedFiles(in_dict)

            if result["OK"]:
                parametersNames = result["Value"]["ParameterNames"]
                records = result["Value"]["Records"]
                if totalrecords == 0:
                    totalrecords = result["Value"]["TotalRecords"]
            else:
                gLogger.error(result["Message"])
                return result

        return S_OK(
            {
                "TotalRecords": totalrecords,
                "ParameterNames": parametersNames,
                "Records": records,
                "Extras": {
                    "Selection": selection,
                    "GlobalStatistics": {"Number of Events": nbOfEvents, "Files Size": filesSize, "Luminosity": lumi},
                },
            }
        )

    #############################################################################
    def getAncestors(self, files, depth):
        """ancestors of given files."""
        return self.db_.getFileAncestors(files, depth)

    #############################################################################
    def getFileCreationLog(self, filename):
        """log file."""
        return self.db_.getFileCreationLog(filename)

    #############################################################################
    def writePythonOrJobOptions(self, startItem, maxitems, path, savetype):
        """create Gaudi Card."""
        result = None
        dataset = None
        if self.parameter_ == self.__bookkeepingParameters[0]:
            result = self._getLimitedFilesConfigParams({"fullpath": path}, {}, startItem, maxitems)
            dataset = self._getDataSetTree1({"fullpath": path})
        elif self.parameter_ == self.__bookkeepingParameters[1]:
            result = self._getLimitedFilesEventTypeParams({"fullpath": path}, {}, startItem, maxitems)
            dataset = self._getDataSetTree2({"fullpath": path})
        elif self.parameter_ == self.__bookkeepingParameters[2]:
            result = self._getLimitedFilesProductions({"fullpath": path}, {}, startItem, maxitems)
            dataset = self._getDataSetTree3({"fullpath": path})
        elif self.parameter_ == self.__bookkeepingParameters[3]:
            result = self._getLimitedFilesRuns({"fullpath": path}, {}, startItem, maxitems)
            dataset = self._getDataSetTree4({"fullpath": path})

        if "TotalRecords" in result["Value"] and result["Value"]["TotalRecords"] > 0:
            records = result["Value"]["Records"]
            params = result["Value"]["ParameterNames"]
            # The list has to be converted to dictionary
            nameIndex = params.index("Name")
            files = {}
            for rec in records:
                lfn = rec[nameIndex]
                files[lfn] = dict(zip(params, rec))
            return self.writeJobOptions(
                files, optionsFile="", savedType=savetype, catalog=None, savePfn=None, dataset=dataset
            )
        else:
            return S_ERROR("Error discovered during the option file creation!")

    #############################################################################
    def getLimitedInformations(self, startItem, maxitems, path):
        """statistics."""
        result = None
        if self.parameter_ == self.__bookkeepingParameters[0]:
            result = self._getLimitedFilesConfigParams({"fullpath": path}, {"need": 0}, startItem, maxitems)
        elif self.parameter_ == self.__bookkeepingParameters[1]:
            result = self._getLimitedFilesEventTypeParams({"fullpath": path}, {"need": 0}, startItem, maxitems)
        elif self.parameter_ == self.__bookkeepingParameters[2]:
            result = self._getLimitedFilesProductions({"fullpath": path}, {"need": 0}, startItem, maxitems)
        elif self.parameter_ == self.__bookkeepingParameters[3]:
            result = self._getLimitedFilesRuns({"fullpath": path}, {"need": 0}, startItem, maxitems)

        if "TotalRecords" in result["Value"] and result["Value"]["TotalRecords"] > 0:
            nbe = 0
            fsize = 0
            nbfiles = 0
            params = result["Value"]["ParameterNames"]
            for lfn in result["Value"]["Records"]:
                nbfiles += 1
                if lfn[params.index("EventStat")] is not None:
                    nbe += int(lfn[params.index("EventStat")])
                if lfn[params.index("FileSize")] is not None:
                    fsize += int(lfn[params.index("FileSize")])
            return S_OK({"Number of Events": nbe, "Files Size": fsize, "Number of files": nbfiles})
        else:
            return S_ERROR("Error discoverd during the option file creation!")

    #############################################################################
    def writeJobOptions(self, files, optionsFile=None, savedType=None, catalog=None, savePfn=None, dataset=None):
        """create options file.

        :params dict files: LFNs for which to write the options
        :params str optionsFile: file name where options might be written (iff given)
        :params str savedType: if savedType == 'txt', return the list of LFNs
        :params str catalog: add catalog or not
        :params dict savePfn: file type versions
        :params dict dataset: metadata about the dataset

        :returns: str with options
        """
        if savedType == "py" or savedType is None:
            # get list of event types
            evtTypes = self.__createEventTypeList(files)

            string = self.__addGaudiheader(evtTypes)
            string += self.__addDatasetCreationMetadata(dataset)

            filesandformats = self.__getFilesandFormats(savePfn, files)
            string += self.__createFormatString(filesandformats)

            if catalog:
                string += "\nFileCatalog().Catalogs += [ 'xmlcatalog_file:" + os.path.basename(catalog) + "' ]\n"
        elif savedType == "txt":
            # Only return the list of LFNs
            string = "\n".join(str(lfn) for lfn in files) + "\n"
        else:
            raise NotImplementedError(savedType)

        if optionsFile:
            # Write options file if requested
            with open(optionsFile, "w") as fd:
                fd.write(string)

        # Always return the string
        return string

    #############################################################################
    def __addDatasetCreationMetadata(self, dataset):
        """it adds the metadata information about the dataset creation."""
        string = ""
        if dataset:
            string += f"\n{self.comment} Extra information about the data processing phases:\n"
            retVal = self.db_.getStepsMetadata(dataset)
            if retVal["OK"]:
                for ppass, record in retVal["Value"]["Records"].items():
                    ppass = dataset.get("ProcessingPass", ppass)
                    string += f"\n{self.comment} Processing Pass: '{ppass}' \n\n"
                    for i in record:
                        string += f"{self.comment} {i[0]} : {i[1]} \n"
        return string

    #############################################################################
    @staticmethod
    def __createFormatString(filesandformats):
        """It generates the Root format option file."""
        string = "\nfrom Gaudi.Configuration import * "
        string += "\nfrom GaudiConf import IOHelper\n"
        for fileFormat, lfns in filesandformats.items():
            if fileFormat:
                string += f"IOHelper('{fileFormat}').inputFiles([\n"
            else:
                string += "IOHelper().inputFiles([\n"
            string += "\n".join(f"'LFN:{lfn}'," for lfn in lfns)
            string += "\n], clear=True)\n"

        return string

    #############################################################################
    def __addGaudiheader(self, evtTypes):
        """it creates the header of the job option."""
        string = self.comment + "GAUDI jobOptions generated on " + time.asctime() + "\n"
        if evtTypes:
            string += self.comment + "Contains event types : \n"
        for evtType in sorted(evtTypes):
            string += self.comment + "  %8d - %d files - %d events - %.2f GBytes\n" % (
                evtType,
                evtTypes[evtType][0],
                evtTypes[evtType][1],
                evtTypes[evtType][2],
            )
        return string

    #############################################################################
    def __getFilesandFormats(self, savePfn, files):
        """It returns a list of lfns for each file format (as a dictionary)"""
        filesandformats = {}
        if savePfn:
            # we have to decide the file type version.
            # This variable contains the file type version, if it is empty I check in the bkk
            lfns = []
            for lfn in files:
                if lfn not in savePfn:
                    lfns.append(lfn)
                    continue
                if isinstance(savePfn[lfn], list):
                    fileFormat = savePfn[lfn][0]["pfntype"]
                else:
                    fileFormat = savePfn[lfn]["pfntype"]
                filesandformats.setdefault(fileFormat, []).append(lfn)

        else:
            lfns = list(files)

        if lfns:
            # Get file type version from BK
            retVal = self.db_.getFileTypeVersion(lfns)
            if retVal["OK"]:
                for lfn, fileFormat in retVal["Value"].items():
                    filesandformats.setdefault(fileFormat, []).append(lfn)
                    lfns.remove(lfn)
            # If no persistency is found, set it to None
            if lfns:
                filesandformats[None] = lfns
        return filesandformats

    #############################################################################
    @staticmethod
    def __createEventTypeList(files):
        """It creates a dictionary which contains the event types and the size of
        the data set."""
        evtTypes = {}
        if not isinstance(files, dict):
            return evtTypes
        for metadata in files.values():
            evtType = metadata.get("EventType")
            if evtType:
                try:
                    stat = int(metadata["EventStat"])
                except (KeyError, ValueError, TypeError):
                    stat = 0

                info = evtTypes.setdefault(int(evtType), [0, 0, 0.0])
                info[0] += 1
                info[1] += stat
                size = metadata.get("FileSize")
                if size is not None:
                    info[2] += int(size) / 1e9
        return evtTypes

    #############################################################################
    def getProcessingPassSteps(self, in_dict):
        """steps for a given processing pass."""
        return self.db_.getProcessingPassSteps(in_dict)

    #############################################################################
    def getMoreProductionInformations(self, prodid):
        """production statistics."""
        return self.db_.getMoreProductionInformations(prodid)

    #############################################################################
    def getAvailableProductions(self):
        """available productions."""
        return self.db_.getAvailableProductions()

    #############################################################################
    def getFileHistory(self, lfn):
        """history."""
        return self.db_.getFileHistory(lfn)

    #############################################################################
    def getProductionProcessingPassSteps(self, in_dict):
        """step of a production."""
        return self.db_.getProductionProcessingPassSteps(in_dict)

    #############################################################################
    def getAvailableDataQuality(self):
        """all the existing data qualities."""
        return self.db_.getAvailableDataQuality()

    #############################################################################
    def setDataQualities(self, values):
        """setting data quality."""
        self.dataQualities_ = values

    def __getSelectedQualities(self):
        """data quality."""
        return [flag for flag, val in self.dataQualities_.items() if val]

    #############################################################################
    def getStepsMetadata(self, bkDict):
        """It is a wrapper to the bookkeeping client."""
        return self.db_.getStepsMetadata(bkDict)

    #############################################################################
    def getFilesWithMetadata(self, dataset):
        """it sets the file types.

        :param dict dataset: it is a bookkeeping dictionary, which contains the conditions used to retreive the lfns
        :return: S_OK lfns with metadata
        """
        return self.db_.getFilesWithMetadata(dataset)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC.AccountingSystem.private.Plotters.DataStoragePlotter.

DataStoragePlotter.__bases__:
  DIRAC.AccountingSystem.private.Plotters.BaseReporter.BaseReporter
"""
from DIRAC import S_OK, S_ERROR
from DIRAC.AccountingSystem.private.Plotters.BaseReporter import BaseReporter

from LHCbDIRAC.AccountingSystem.Client.Types.DataStorage import DataStorage

# FIXME: refactor _reportMethods
# FIXME: refactor _plotMethods


class DataStoragePlotter(BaseReporter):
    """DataStoragePlotter as extension of BaseReporter."""

    _typeName = "DataStorage"
    _typeKeyFields = [dF[0] for dF in DataStorage().definitionKeyFields]

    # Catalog Space

    _reportCatalogSpaceName = "LFN size"

    def _reportCatalogSpace(self, reportRequest):
        """Reports about the LFN size and the catalog space from the accounting.

        :param reportRequest: <dict>
          { 'grouping'       : 'EventType',
            'groupingFields' : ( '%s', [ 'EventType' ] ),
            'startTime'      : 1355663249.0,
            'endTime'        : 1355749690.0,
            'condDict'       : { 'EventType' : 'Full stream' }
          }

        returns S_OK / S_ERROR
          { 'graphDataDict': { 'Full stream': { 1355616000L: 935.38852424691629,
                                                1355702400L: 843.84448707482204 }
                             },
            'data'         : { 'Full stream': { 1355616000L: 935388.52424691629,
                                                1355702400L: 843844.48707482207 }
                             },
            'unit'         : 'GB',
            'granularity'  : 86400
          }
        """

        if reportRequest["grouping"] == "StorageElement":
            return S_ERROR("Grouping by storage element when requesting lfn info makes no sense")

        selectString = self._getSelectStringForGrouping(reportRequest["groupingFields"])

        selectFields = (
            selectString + ", %s, %s, SUM(%s/%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "LogicalSize", "entriesInBucket"],
        )

        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {"convertToGranularity": "average", "checkNone": True},
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)

        accumMaxValue = self._getAccumulationMaxValue(dataDict)
        suitableUnits = self._findSuitableUnit(dataDict, accumMaxValue, "bytes")

        # 3rd variable unused ( maxValue )
        baseDataDict, graphDataDict, __, unitName = suitableUnits

        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotCatalogSpace(self, reportRequest, plotInfo, filename):
        """Creates <filename>.png file containing information regarding the LFN
        size and the catalog space.

        :param reportRequest: <dict>
           { 'grouping'       : 'EventType',
             'groupingFields' : ( '%s', [ 'EventType' ] ),
             'startTime'      : 1355663249.0,
             'endTime'        : 1355749690.0,
             'condDict'       : { 'EventType' : 'Full stream' }
           }
        :param plotInfo: <dict> ( output of _reportCatalogSpace )
           { 'graphDataDict' : { 'Full stream' : { 1355616000L: 4.9003546130956819,
                                                   1355702400L: 4.9050379437065059 }
                               },
             'data'          : { 'Full stream' : { 1355616000L: 4900354613.0956821,
                                                   1355702400L: 4905037943.7065058 }
                               },
             'unit'          : 'PB',
             'granularity'   : 86400
            }
        :param filename: <str>
          '_plotCatalogSpace'

        returns S_OK / S_ERROR
           { 'plot': True, 'thumbnail': False }
        """

        startTime = reportRequest["startTime"]
        endTime = reportRequest["endTime"]
        span = plotInfo["granularity"]
        dataDict = plotInfo["graphDataDict"]

        metadata = {
            "title": f"LFN space usage grouped by {reportRequest['grouping']}",
            "starttime": startTime,
            "endtime": endTime,
            "span": span,
            "ylabel": plotInfo["unit"],
        }

        dataDict = self._fillWithZero(span, startTime, endTime, dataDict)
        return self._generateStackedLinePlot(filename, dataDict, metadata)

    # .............................................................................
    # Catalog Files

    _reportCatalogFilesName = "LFN files"

    def _reportCatalogFiles(self, reportRequest):
        """Reports about the LFN files and the catalog files from the accounting.

        :param reportRequest: <dict>
          { 'grouping'       : 'EventType',
            'groupingFields' : ( '%s', [ 'EventType' ] ),
            'startTime'      : 1355663249.0,
            'endTime'        : 1355749690.0,
            'condDict'       : { 'EventType' : 'Full stream' }
          }

        returns S_OK / S_ERROR
          { 'graphDataDict' : { 'Full stream' : { 1355616000L : 420.47885754501203,
                                                  1355702400L : 380.35170637810842 }
                                                },
            'data'          : { 'Full stream' : { 1355616000L : 420.47885754501203,
                                                  1355702400L : 380.35170637810842 }
                                                },
            'unit'          : 'files',
            'granularity'   : 86400
          }
        """

        if reportRequest["grouping"] == "StorageElement":
            return S_ERROR("Grouping by storage element when requesting lfn info makes no sense")

        selectString = self._getSelectStringForGrouping(reportRequest["groupingFields"])
        selectFields = (
            selectString + ", %s, %s, SUM(%s/%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "LogicalFiles", "entriesInBucket"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {"convertToGranularity": "average", "checkNone": True},
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)

        accumMaxValue = self._getAccumulationMaxValue(dataDict)
        suitableUnits = self._findSuitableUnit(dataDict, accumMaxValue, "files")

        # 3rd variable unused ( maxValue )
        baseDataDict, graphDataDict, __, unitName = suitableUnits

        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotCatalogFiles(self, reportRequest, plotInfo, filename):
        """Creates <filename>.png file containing information regarding the LFN
        files and the catalog files.

        :param reportRequest: <dict>
          { 'grouping'       : 'EventType',
            'groupingFields' : ( '%s', [ 'EventType' ] ),
            'startTime'      : 1355663249.0,
            'endTime'        : 1355749690.0,
            'condDict'       : { 'EventType' : 'Full stream' }
          }
        :param plotInfo: <dict> ( output of _reportCatalogFiles )
          { 'graphDataDict' : { 'Full stream' : { 1355616000L : 420.47885754501203,
                                                  1355702400L : 380.35170637810842 }
                              },
            'data'          : { 'Full stream' : { 1355616000L : 420.47885754501203,
                                                  1355702400L : 380.35170637810842 }
                              },
            'unit'          : 'files',
            'granularity'   : 86400
          }
        :param filename: <str>
          '_plotCatalogFiles'

        returns S_OK / S_ERROR
           { 'plot': True, 'thumbnail': False }
        """

        startTime = reportRequest["startTime"]
        endTime = reportRequest["endTime"]
        span = plotInfo["granularity"]
        dataDict = plotInfo["graphDataDict"]

        metadata = {
            "title": f"Number of LFNs by {reportRequest['grouping']}",
            "starttime": startTime,
            "endtime": endTime,
            "span": span,
            "ylabel": plotInfo["unit"],
        }

        dataDict = self._fillWithZero(span, startTime, endTime, dataDict)
        return self._generateStackedLinePlot(filename, dataDict, metadata)

    # .............................................................................
    # Physical Space

    _reportPhysicalSpaceName = "PFN size"

    def _reportPhysicalSpace(self, reportRequest):
        """Reports about the PFN size and the physical space from the accounting.

        :param reportRequest: <dict>
          { 'grouping'       : 'EventType',
            'groupingFields' : ( '%s', [ 'EventType' ] ),
            'startTime'      : 1355663249.0,
            'endTime'        : 1355749690.0,
            'condDict'       : { 'EventType' : 'Full stream' }
          }

        returns S_OK / S_ERROR
          { 'graphDataDict' : { 'Full stream' : { 1355616000L : 14.754501202,
                                                  1355702400L : 15.237810842 }
                              },
            'data'          : { 'Full stream' : { 1355616000L : 14.754501202,
                                                  1355702400L : 15.237810842 }
                              },
            'unit'          : 'MB',
            'granularity'   : 86400
          }
        """

        selectString = self._getSelectStringForGrouping(reportRequest["groupingFields"])
        selectFields = (
            selectString + ", %s, %s, SUM(%s/%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "PhysicalSize", "entriesInBucket"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {"convertToGranularity": "average", "checkNone": True},
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)

        accumMaxValue = self._getAccumulationMaxValue(dataDict)
        suitableUnits = self._findSuitableUnit(dataDict, accumMaxValue, "bytes")

        # 3rd variable unused ( maxValue )
        baseDataDict, graphDataDict, __, unitName = suitableUnits

        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotPhysicalSpace(self, reportRequest, plotInfo, filename):
        """Creates <filename>.png file containing information regarding the PFN
        size and the physical space.

        :param reportRequest: <dict>
          { 'grouping'       : 'EventType',
            'groupingFields' : ( '%s', [ 'EventType' ] ),
            'startTime'      : 1355663249.0,
            'endTime'        : 1355749690.0,
            'condDict'       : { 'EventType' : 'Full stream' }
          }
        :param plotInfo: <dict> ( output of _reportPhysicalSpace )
          { 'graphDataDict' : { 'Full stream' : { 1355616000L : 14.754501202,
                                                  1355702400L : 15.237810842 }
                              },
            'data'          : { 'Full stream' : { 1355616000L : 14.754501202,
                                                  1355702400L : 15.237810842 }
                              },
            'unit'          : 'MB',
            'granularity'   : 86400
          }
        :param filename: <str>
          '_plotPhysicalSpace'

        returns S_OK / S_ERROR
           { 'plot': True, 'thumbnail': False }
        """

        startTime = reportRequest["startTime"]
        endTime = reportRequest["endTime"]
        span = plotInfo["granularity"]
        dataDict = plotInfo["graphDataDict"]

        metadata = {
            "title": f"PFN space usage by {reportRequest['grouping']}",
            "starttime": startTime,
            "endtime": endTime,
            "span": span,
            "ylabel": plotInfo["unit"],
        }

        dataDict = self._fillWithZero(span, startTime, endTime, dataDict)
        return self._generateStackedLinePlot(filename, dataDict, metadata)

    # .............................................................................
    # Physical Files

    _reportPhysicalFilesName = "PFN files"

    def _reportPhysicalFiles(self, reportRequest):
        """Reports about the PFN files and the physical files from the accounting.

        :param reportRequest: <dict>
           { 'grouping'       : 'EventType',
             'groupingFields' : ( '%s', [ 'EventType' ] ),
             'startTime'      : 1355663249.0,
             'endTime'        : 1355749690.0,
             'condDict'       : { 'EventType' : 'Full stream' }
           }

        returns S_OK / S_ERROR
          { 'graphDataDict' : { 'Full stream' : { 1355616000L : 42.47885754501203,
                                                  1355702400L : 38.35170637810842 }
                              },
            'data'          : { 'Full stream' : { 1355616000L : 42.47885754501203,
                                                  1355702400L : 38.35170637810842 }
                              },
            'unit'          : 'files',
            'granularity'   : 86400
          }
        """

        selectString = self._getSelectStringForGrouping(reportRequest["groupingFields"])
        selectFields = (
            selectString + ", %s, %s, SUM(%s/%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "PhysicalFiles", "entriesInBucket"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {"convertToGranularity": "average", "checkNone": True},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)

        accumMaxValue = self._getAccumulationMaxValue(dataDict)
        suitableUnits = self._findSuitableUnit(dataDict, accumMaxValue, "files")

        # 3rd variable unused ( maxValue )
        baseDataDict, graphDataDict, __, unitName = suitableUnits

        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotPhysicalFiles(self, reportRequest, plotInfo, filename):
        """Creates <filename>.png file containing information regarding the PFN
        files and the physical files.

        :param reportRequest: <dict>
           { 'grouping'       : 'EventType',
             'groupingFields' : ( '%s', [ 'EventType' ] ),
             'startTime'      : 1355663249.0,
             'endTime'        : 1355749690.0,
             'condDict'       : { 'EventType' : 'Full stream' }
           }
        :param plotInfo: <dict> ( output of _reportPhysicalFiles )
          { 'graphDataDict' : { 'Full stream' : { 1355616000L: 4.9003546130956819,
                                                  1355702400L: 4.9050379437065059 }
                              },
            'data'          : { 'Full stream' : { 1355616000L: 4900354613.0956821,
                                                  1355702400L: 4905037943.7065058 }
                              },
            'unit'          : 'PB',
            'granularity'   : 86400
          }
        :param filename: <str>
          '_plotPhysicalFiles'

        return S_OK / S_ERROR
           { 'plot': True, 'thumbnail': False }
        """

        startTime = reportRequest["startTime"]
        endTime = reportRequest["endTime"]
        span = plotInfo["granularity"]
        dataDict = plotInfo["graphDataDict"]

        metadata = {
            "title": f"Number of PFNs by {reportRequest['grouping']}",
            "starttime": startTime,
            "endtime": endTime,
            "span": span,
            "ylabel": plotInfo["unit"],
        }

        dataDict = self._fillWithZero(span, startTime, endTime, dataDict)
        return self._generateStackedLinePlot(filename, dataDict, metadata)



#! /usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""In a transformation, flush a list of runs or runs that are flushed in the
transformation used in BKQuery."""
import DIRAC
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from DIRAC import gLogger
    from DIRAC.TransformationSystem.Utilities.ScriptUtilities import getTransformations

    Script.registerSwitch("", "Runs=", "   list of runs to flush (comma separated, ranges r1:r2)")
    Script.registerSwitch("", "NoAction", "   No action taken, just give stats")
    Script.registerSwitch("", "Active", "   If used, selects all active runs")
    Script.setUsageMessage(
        "\n".join(
            [
                __doc__,
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] <TransID>[,<TransID2>...]",
            ]
        )
    )
    Script.parseCommandLine()

    action = True
    runList = []
    active = False
    for switch, val in Script.getUnprocessedSwitches():
        if switch == "NoAction":
            action = False
        elif switch == "Active":
            active = True
        elif switch == "Runs":
            try:
                runs = val.split(",")
                for run in runs:
                    if run.isdigit():
                        runList.append(int(run))
                    else:
                        runRange = run.split(":")
                        if len(runRange) == 2 and runRange[0].isdigit() and runRange[1].isdigit():
                            runList += range(int(runRange[0]), int(runRange[1]) + 1)
            except Exception as x:
                gLogger.exception("Bad run parameter", lException=x)

    transList = getTransformations(Script.getPositionalArgs())
    if not transList:
        DIRAC.exit(1)
    from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

    transClient = TransformationClient()

    for transID in transList:
        res = transClient.getTransformationRuns({"TransformationID": transID})
        if not res["OK"]:
            gLogger.fatal(f"Error getting runs for transformation {transID}", res["Message"])
            DIRAC.exit(1)
        runs = res["Value"]
        runs.sort(key=lambda x: int(x["RunNumber"]))

        if not runList:
            if active:
                # Flush all Active runs in that transformation
                toBeFlushed = [run["RunNumber"] for run in runs if run["Status"] != "Flush"]
            else:
                # Get transformationID used in the BK query, to find out which runs are flushed in there
                res = transClient.getBookkeepingQuery(transID)
                if not res["OK"]:
                    gLogger.fatal(f"Error getting BK query for transformation {transID}", res["Message"])
                    DIRAC.exit(1)
                queryProd = res["Value"].get("ProductionID")
                if not queryProd:
                    gLogger.fatal("Transformation is not based on another one")
                    DIRAC.exit(0)

                res = transClient.getTransformationRuns({"TransformationID": queryProd})
                if not res["OK"]:
                    gLogger.fatal(f"Error getting runs for transformation {queryProd}", res["Message"])
                    DIRAC.exit(1)
                queryRuns = res["Value"]
                # Get the list of runs flushed in the parent production
                flushedRuns = [run["RunNumber"] for run in queryRuns if run["Status"] == "Flush"]

                toBeFlushed = []
                # Get runs not flushed in current production but flushed in the parent production
                for run in [
                    run["RunNumber"] for run in runs if run["Status"] != "Flush" and run["RunNumber"] in flushedRuns
                ]:
                    missing = -1
                    res = transClient.getTransformationFiles({"TransformationID": queryProd, "RunNumber": run})
                    if not res["OK"]:
                        gLogger.fatal("Error getting files for run %d" % run, res["Message"])
                    else:
                        runFiles = res["Value"]
                        missing = 0
                        for rFile in runFiles:
                            if rFile["Status"] in ("Unused", "Assigned", "MaxReset"):
                                missing += 1
                        if not missing:
                            toBeFlushed.append(run)

        else:
            toBeFlushed = sorted(set(runList) & {run["RunNumber"] for run in runs})

        ok = 0
        gLogger.always(
            f"Runs {'being' if action else 'to be'} flushed in transformation {transID}:",
            ",".join([str(run) for run in toBeFlushed]),
        )
        for run in toBeFlushed:
            res = transClient.setTransformationRunStatus(transID, run, "Flush") if action else {"OK": True}
            if not res["OK"]:
                gLogger.fatal(f"Error setting run {run} to Flush in transformation {transID}", res["Message"])
            else:
                ok += 1
        gLogger.always("%d runs set to Flush in transformation %d" % (ok, transID))


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import math
import sys
from collections import defaultdict
from datetime import datetime, timezone

from rich.table import Table
from rich.markdown import Markdown
from rich.prompt import Prompt

from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

from .actions import (
    ExtendEndRunAction,
    SubmitProductionsAction,
    UpdateIssueStateLabel,
    CheckValidationAction,
    CleanValidationAction,
    CompleteTransformationAction,
    CloseIssueAction,
    AddLabelAction,
    UpdateTransformationStatusAction,
    UpdateRequestStateAction,
    SubmitRemovalTransformationAction,
)
from .integrations import EVENTTYPE_TO_STREAM

ACTIVE_PRODUCTION_STATES = {
    "run-validation",
    "ready",
    "running",
    "running-concurrent",
    "update-end-run",
    "checking",
    "debugging",
}
CHECK_DISPLAY = {None: "❔", True: "✅"}


def status_emoji(status):
    if status in {"Archived", "Completed"}:
        return ""
    if status in {"Cleaned", "Cleaning"}:
        return "🧹"
    if status in {"Active", "Idle"}:
        return "🏃"
    if status in {"Stopped"}:
        return "⏹️"
    return "❔"


def analyse_active_productions(repo, *, states: set[str] = ACTIVE_PRODUCTION_STATES):
    all_issues = repo.poll(states=states)
    actions = defaultdict(list)

    # Find validations that need to be submitted
    for issue in all_issues.get("run-validation", []):
        needs_submit = True
        for request_id, meta in issue.metadata.validations.items():
            if meta.running:
                actions[issue].append(CheckValidationAction(issue, request_id))
                needs_submit = False
        if needs_submit:
            actions[issue].append(SubmitProductionsAction(issue))

    # Find productions that need to be submitted
    for issue in all_issues.get("ready", []):
        for request_id, meta in issue.metadata.validations.items():
            if not meta.cleaned:
                actions[issue].append(CleanValidationAction(issue, request_id))
        actions[issue].append(SubmitProductionsAction(issue))

    issues_by_state = {k: v for k, v in all_issues.items() if k in states}
    # Find metadata for all relevant transformations
    all_tids = set()
    for state, issues in issues_by_state.items():
        for issue in issues:
            for request_id, request_meta in issue.metadata.requests.items():
                all_tids.update(request_meta.all_transform_ids)
    if all_tids:
        retVal = TransformationClient().getTransformations(
            {"TransformationID": list(all_tids)},
            columns=[
                "TransformationFamily",
                "TransformationID",
                "Type",
                "TransformationGroup",
                "CreationDate",
                "Status",
            ],
        )
        tinfo = {x["TransformationID"]: x for x in returnValueOrRaise(retVal)}
        input_queries = returnValueOrRaise(TransformationClient().getBookkeepingQueries(list(tinfo)))
    else:
        tinfo = {}
        input_queries = {}

    # Create the table data
    last_update = datetime.fromtimestamp(0, timezone.utc)
    tables_data = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(list))))
    for state, issues in issues_by_state.items():
        for issue in issues:
            last_update = max(last_update, issue.metadata.last_updated)
            for request_id, request_meta in issue.metadata.requests.items():
                main_tid, merge_tid = request_meta.transform_ids
                if tinfo[merge_tid]["Type"] != "Merge":
                    raise NotImplementedError(tinfo[merge_tid])
                input_query = input_queries[main_tid]
                proc_pass = tinfo[main_tid]["TransformationGroup"].split("/")[-1]
                row_data = [f"[link={issue.url}]#{issue.issue.iid}[/link]"]
                row_data += [request_id, state]
                row_data += [input_query["DataTakingConditions"]]
                row_data += [
                    f"{tid} {status_emoji(tinfo[tid]['Status'])}" if tid else "⚠️"
                    for tid in [main_tid, merge_tid, request_meta.removal, request_meta.replication]
                ]
                row_data += [f"{input_query['StartRun']}:{input_query['EndRun']}"]
                file_status = request_meta.file_status.get(main_tid, {})
                total = sum(file_status.values())
                row_data.append(total)
                row_data.append(file_status.get("Processed", 0))
                for status in ["MaxReset", "Problematic"]:
                    count = file_status.get(status, 0)
                    row_data.append(f"{count} 🚨" if count else 0)
                for file_status_tid in [main_tid, merge_tid, request_meta.removal]:
                    file_status = request_meta.file_status.get(file_status_tid, {})
                    total = sum(file_status.values())
                    row_data.append(
                        f"{math.floor(file_status.get('Processed', 0) / total * 100_00) / 100:.2f}%" if total else ""
                    )
                n_checks = 0
                n_checks_passed = 0
                for check_name in ["dm_check", "pm_check"]:
                    check_status = [getattr(request_meta.checks, check_name).get(x) for x in request_meta.transform_ids]
                    row_data.append(" ".join(CHECK_DISPLAY[x] for x in check_status))
                    n_checks += len(check_status)
                    n_checks_passed += sum(filter(None, check_status))
                req_cleaned = request_meta.checks.dm_clean.get(request_id)
                row_data.append(CHECK_DISPLAY[req_cleaned])
                n_checks += 1
                n_checks_passed += bool(req_cleaned)

                config = input_query["ConfigVersion"]
                event_type = input_query["EventType"]
                expected_labels = {f"config::{config}", f"proc::{proc_pass}", f"stream::{event_type}"}
                for label in expected_labels:
                    if label not in issue.labels:
                        actions[issue].append(AddLabelAction(repo, issue, label))

                tables_data[config][proc_pass][event_type][tinfo[main_tid]["Type"]].append(row_data)

                if state == "running":
                    if not request_meta.removal:
                        actions[issue].append(SubmitRemovalTransformationAction(issue, request_id))

                    all_finished = True
                    for tid in request_meta.transform_ids + [request_meta.removal]:
                        file_status = request_meta.file_status.get(tid, {})
                        total = sum(file_status.values())
                        finished = sum(
                            file_status.get(status, 0) for status in ["Processed", "NotProcessed", "Removed"]
                        )
                        if total == 0 or total != finished:
                            all_finished = False
                            continue
                        if tid not in tinfo:
                            print(f"Missing transformation {tid}, maybe a metadata refresh is needed?")
                            continue
                        match tinfo[tid]["Status"]:
                            case "Active":
                                all_finished = False
                            case "Idle":
                                actions[issue].append(UpdateTransformationStatusAction(tid, "Stopped"))
                    if all_finished:
                        actions[issue].append(UpdateIssueStateLabel(issue, "running", "checking"))

                if state == "checking" and n_checks == n_checks_passed:
                    actions[issue].append(CompleteTransformationAction(issue, main_tid))
                    actions[issue].append(CompleteTransformationAction(issue, merge_tid))
                    actions[issue].append(CompleteTransformationAction(issue, request_meta.removal))
                    actions[issue].append(UpdateRequestStateAction(request_id, "Done"))
                    actions[issue].append(UpdateIssueStateLabel(issue, "checking", "done"))
                    actions[issue].append(CloseIssueAction(issue))

                # Ensure the start run is sensible
                if input_queries.get(main_tid, {}).get("StartRun") != issue.run_yaml.get("start_run"):
                    raise NotImplementedError(f"Start run mismatch: {issue} {main_tid}")
                if (
                    request_meta.removal
                    and input_queries.get(request_meta.removal, {}).get("StartRun")
                    != input_queries[main_tid]["StartRun"]
                ):
                    raise NotImplementedError(f"Removal start run mismatch: {issue} {request_meta.removal}")

                # Check if the end run needs to be extended
                main_value = input_queries.get(main_tid, {}).get("EndRun")
                expected_value = issue.run_yaml.get("end_run")
                if main_value != expected_value:
                    if state != "update-end-run":
                        raise NotImplementedError(
                            f"End run mismatch: {issue} {main_tid} {main_value} != {expected_value}"
                        )
                    actions[issue].append(ExtendEndRunAction(issue, main_tid, expected_value))
                    if request_meta.removal is not None:
                        actions[issue].append(ExtendEndRunAction(issue, request_meta.removal, expected_value))
                    new_state = "running-concurrent" if issue.run_yaml["concurrent"] else "running"
                    actions[issue].append(UpdateIssueStateLabel(issue, "update-end-run", new_state))
                elif state == "update-end-run":
                    raise NotImplementedError(f"State is update-end-run there is nothing to change: {issue}")

                # Ensure the removal is consistent with the main transformation
                if request_meta.removal is not None:
                    removal_value = input_queries.get(request_meta.removal, {}).get("EndRun")
                    if removal_value != main_value:
                        raise NotImplementedError(
                            f"Removal end run mismatch: {issue} {request_meta.removal} {removal_value} != {main_value}"
                        )

    return last_update, tables_data, actions


def display_table(console, tables_data):
    # Create the rich tables from the table data
    tables = defaultdict(list)
    for config in sorted(tables_data):
        for proc_pass in sorted(tables_data[config]):
            section = f"# {config} {proc_pass}"
            for eventtype in sorted(tables_data[config][proc_pass]):
                for tran_type, table_rows in sorted(tables_data[config][proc_pass][eventtype].items()):
                    table = Table(
                        *["Issue", "Request", "State", "Conditions", tran_type, "Merging", "Removal", "Replication"],
                        *["Runs", "RAWs", "Processed", "MaxReset", "Problematic", "% Done", "(merge)", "(removal)"],
                        *["DM Check", "PM Check", "DM Clean"],
                        row_styles=["dim", ""],
                        title=f"[bold]{eventtype} ({EVENTTYPE_TO_STREAM.get(eventtype)})",
                    )
                    for row in sorted(table_rows, key=lambda x: x[4]):
                        table.add_row(*map(str, row))
                    tables[section].append(table)

    # Make the columns the same width across all tables
    for cols in zip(*(x.columns for section_tables in tables.values() for x in section_tables)):
        width = max(console.measure(cell).maximum for col in cols for cell in col.cells)
        width = max(width, len(cols[0].header))
        for col in cols:
            col.width = width
    # Display the tables
    for section, tables in tables.items():
        console.print(Markdown(section))
        for table in tables:
            console.print(table)


def display_actions(console, actions, *, execute=False):
    console.print(Markdown("# Recommended Actions"))
    for issue, action_list in actions.items():
        lines = [f"## {issue.url} ({issue.state})\n"]

        automatic_actions = []
        prompted_actions = []
        for action in action_list:
            if action.safe:
                automatic_actions.append(action)
            else:
                prompted_actions.append(action)

        if automatic_actions:
            lines.append("\nAutomatic actions:\n")
            lines.extend(f"- {action.message()}" for action in automatic_actions)

        if prompted_actions:
            lines.append("\nManual actions:\n")
            lines.extend(f"- {action.message()}" for action in prompted_actions)

        console.print(Markdown("\n".join(lines)))

        if not execute:
            continue

        for action in automatic_actions:
            action.run()

        if not prompted_actions:
            continue

        match response := Prompt.ask(
            "Do you want to run these manual actions?", choices=["run-all", "ask-each", "no", "quit"], default="no"
        ):
            case "run-all":
                for action in prompted_actions:
                    action.run()
            case "ask-each":
                for action in prompted_actions:
                    if Prompt.ask(f"Run the action: {action.message()}?", choices=["yes", "no"]) == "yes":
                        action.run()
            case "no":
                continue
            case "quit":
                sys.exit(0)
            case _:
                raise NotImplementedError(f"Invalid response: {response}")



###############################################################################
# (c) Copyright 2022 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""The HistogramMergingAgent finds MC transformations that are finished with
testing and merges their HIST files for SimDQ monitoring.

An overview of the procedure:
1. take all MCSimulation/MCFastSimulation/MCReconstruction transformations that
have finished testing but are not yet archived,
2. deduce the output LFNs for the merged histograms from each transformation,
3. if the output LFN does not already exist then merge the files,
4. upload and register the new merged files,
5. generate and commit new metadata to the SimProdDB repository.
"""
import hashlib
import re
import shlex
import subprocess
import tempfile
from copy import copy
from datetime import datetime
from functools import lru_cache
from pathlib import Path

import DIRAC
import git
import yaml
from DIRAC import S_OK
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.File import generateGuid
from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise
from DIRAC.Core.Utilities.TimeUtilities import timeThis
from DIRAC.DataManagementSystem.Client.DataManager import DataManager

from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.Core.Utilities.BookkeepingJobInfo import BookkeepingJobInfo
from LHCbDIRAC.Core.Utilities.XMLTreeParser import XMLNode, XMLTreeParser
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

AGENT_NAME = "ProductionManagement/HistogramMergingAgent"

gStandAlone = False  # Do not perform queries to the Bookkeeping/Transformation/ProductionManagement Systems
gLocalWorkDir = False  # Use the CWD for writing files, in case am_getWorkDirectory resolves to a read-only location
gDoRealMerging = True  # Run the `hadd` command to produce actual ROOT files
gDoRealUpdate = True  # OK to upload files, write to Bookkeeping and commit & push to the git repo
gUseGitRepo = True  # OK to check-out and manipulate the SimDQ metadata git repository

REQ_KEYS = {
    # Extra information that Monet can display to the shifter
    # See LHCbDIRAC.ProductionManagementSystem.DB.ProductionRequestDB for full list
    "RequestID",
    "RequestName",
    "RequestWG",
    "SimCondition",
    "SimCondDetail",
    "ProPath",
    "Extra",
    "RetentionRate",
    "FastSimulationType",
}

TFM_KEYS = {
    # Extra information that Monet can display to the shifter
    # See LHCbDIRAC.TransformationSystem.DB.TransformationDB and DIRAC.TransformationSystem.DB.TransformationDB for full list
    "TransformationID",
    "Type",
}


class HistogramMergingAgent(AgentModule):
    """HistogramMergingAgent"""

    def __init__(self, *args, **kwargs):
        """c'tor."""
        super().__init__(*args, **kwargs)
        # Clients
        self.bk_client = None
        self.rp_client = None
        self.tf_client = None
        self.dm_client = None
        self.xml_parser = None
        # Local directories
        self.merge_dir = None
        self.simdq_dir = None
        # Remote targets
        self.simdq_repo = None
        self.hist_se = None
        # Configuration
        self.git_author = git.Actor(
            self.am_getOption("GitAuthorName", "LHCbDIRAC"),
            self.am_getOption("GitAuthorEmail", "lhcb-production-manager@cern.ch"),
        )
        self.ssh_key = self.am_getOption("GitSSHKeyFile", "")
        self.n_cpus = 1  # Passed to `hadd -j`
        if gDoRealUpdate:
            self.simdq_url = "ssh://git@gitlab.cern.ch:7999/lhcb-simulation/simdqdata.git"
        else:
            self.simdq_url = "https://gitlab.cern.ch/lhcb-simulation/simdqdata.git"
        self.simdq_url = self.am_getOption("SimDQRepoURL", self.simdq_url)
        self.simdq_branch = self.am_getOption("SimDQRepoBranch", "main")
        self.lbconda_env_name = self.am_getOption("HistMergeCondaEnv", "default")
        self.allowed_types = self.am_getOption(
            "ProductionTypes", ["MCSimulation", "MCFastSimulation", "MCReconstruction"]
        )
        self.allowed_statuses = self.am_getOption("ProductionStatuses", ["Active", "Idle", "Completed"])
        self.lbconda_env_version = None

    def initialize(self):
        """Set up the clients and merging directory."""
        # Local directories
        if gLocalWorkDir:
            # Useful if am_getWorkDirectory resolves to a read-only file system (e.g. CVMFS)
            work_dir = "."
        else:
            work_dir = self.am_getWorkDirectory()
        self.merge_dir = Path(work_dir) / "MergeHistograms"
        self.simdq_dir = Path(work_dir) / "SimProdDB"
        # Make sure that the directories exist
        for path in [self.merge_dir, self.simdq_dir]:
            path.mkdir(parents=True, exist_ok=True)
        # Remote targets
        self.hist_se = Operations().getValue("Productions/HistogramSE", "CERN-HIST-EOS")

        return S_OK()

    def beginExecution(self):
        """Get latest version of conda env"""
        self.bk_client = BookkeepingClient()
        self.rp_client = ProductionRequestClient()
        self.tf_client = TransformationClient()
        self.xml_parser = XMLTreeParser()
        self.dm_client = DataManager()

        self.lbconda_env_version = self._latest_lbconda_env_version

        return S_OK()

    def execute(self):
        """Main method"""
        # Don't use the server certificate otherwise the DFC wont let us write
        gConfigurationData.setOptionInCFG("/DIRAC/Security/UseServerCertificate", "false")

        transformations = self._fetch_transformations()
        if not transformations:
            # Nothing to do!
            return S_OK()

        if gDoRealUpdate and not Path(self.ssh_key).is_file():
            raise RuntimeError(f"GitSSHKeyFile={self.ssh_key} is not a valid file path")
        self.simdq_repo = self._checkout_metadata()

        request_ids = list({int(t["TransformationFamily"]) for t in transformations})
        requests = self._query_requests(request_ids)

        # Perform the merging and write the metadata files
        for transform in transformations:
            request_id = int(transform["TransformationFamily"])
            request = requests[request_id]
            # Loop over step-filetype combinations
            new_files = []
            for key in transform["HistogramMerging"]:
                new_files += [self._handle_transformation(transform, request, key)]

            # Commit meteadata files to the SimDQ repo
            if gUseGitRepo:
                commit_message = f"add metadata for {len(new_files)} new productions"
                self._commit_and_push_metadata(new_files, commit_message)

        return S_OK()

    # Properties ###############################################################

    @property
    def _latest_lbconda_env_version(self) -> str:
        cmd = ["lb-conda", "--list", self.lbconda_env_name]
        ret = subprocess.run(cmd, check=True, capture_output=True, text=True)
        latest_version = ret.stdout.strip().split("\n")[-1]
        return latest_version

    @property
    @lru_cache
    def _available_hist_filetypes(self) -> list[str]:
        if gStandAlone:
            return [f"{app.upper()}HIST" for app in ["Gauss", "Boole", "Brunel", "Moore", "DaVinci"]]
        all_filetypes = returnValueOrRaise(self.bk_client.getAvailableFileTypes())["Records"]
        hist_filetypes = [filetype for filetype, _ in all_filetypes if filetype.endswith("HIST")]
        return hist_filetypes

    # Main method for obtaining transformations to process #####################

    def _fetch_transformations(self) -> list[dict]:
        """Get recent unprocessed transformations that have histogram files to merge"""
        # Get recent transformations and their additional parameters
        recent_transformations = self._query_recent_transformations()
        for transform in recent_transformations:
            workflow = self._parse_workflow_xml(transform.pop("Body"))
            transform.update({"Parameters": self._get_transformation_parameters(workflow)})
            transform.update({"Steps": self._get_transformation_steps(workflow)})

        # Get histogram file LFNs and corresponding PFNs
        all_transform_ids = [int(t["TransformationID"]) for t in recent_transformations]
        all_lfns = self._query_input_lfns(all_transform_ids)
        if len(all_lfns) == 0:
            return []
        all_pfns = self._query_pfns(all_lfns)
        input_lfn_dict = self._group_lfns_by_prod_and_type(all_lfns)
        input_pfn_dict = {
            transform_id: {filetype: [all_pfns[lfn] for lfn in filetype_dict[filetype]] for filetype in filetype_dict}
            for transform_id, filetype_dict in input_lfn_dict.items()
        }

        # Deduce the corresponding output LFNs for merged hist files
        output_lfn_dict = {
            transform_id: {
                filetype: self._format_output_lfn(
                    [t for t in recent_transformations if int(t["TransformationID"]) == transform_id][0],
                    filetype,
                )
                for filetype in filetype_dict
            }
            for transform_id, filetype_dict in input_lfn_dict.items()
        }
        all_output_lfns = [lfn for filetype_dict in output_lfn_dict.values() for lfn in filetype_dict.values()]
        output_lfn_status = self._query_lfns_exist(all_output_lfns)

        # Update transformation dicts with info about histogram merging
        for transform in recent_transformations:
            if (transform_id := int(transform["TransformationID"])) in input_lfn_dict:
                merge_info = {
                    filetype: {
                        "InputLFNs": input_lfn_dict[transform_id][filetype],
                        "InputPFNs": input_pfn_dict[transform_id][filetype],
                        "OutputLFN": output_lfn_dict[transform_id][filetype],
                    }
                    for filetype in input_lfn_dict[transform_id]
                    if len(input_pfn_dict[transform_id][filetype]) > 0  # There are actually some PFNs to merge
                    and output_lfn_dict[transform_id][filetype]
                    in output_lfn_status["Failed"]  # Output LFN doesn't exist already
                }
                if len(merge_info) > 0:
                    transform.update({"HistogramMerging": merge_info})

        filtered_transformations = filter(lambda t: "HistogramMerging" in t, recent_transformations)

        return list(filtered_transformations)

    # Main method for processing a single transformation #######################

    def _handle_transformation(
        self,
        transform: dict,
        request: dict,
        file_key: str,
    ) -> Path:
        """Perform the merging, upload the file, write metadata to Bookkeeping and SimDQ repo"""

        merge_info = transform["HistogramMerging"][file_key]

        # Deduce filename of merged histogram file
        output_lfn = merge_info["OutputLFN"]
        output_fn = self.merge_dir / output_lfn.rsplit("/", maxsplit=1)[-1]

        # Merge the histogram files locally
        start = datetime.utcnow()
        self._write_merged_hist_file(merge_info["InputPFNs"], output_fn)
        end = datetime.utcnow()

        # Get some metadata about the merged histogram file
        output_md5 = hashlib.md5(output_fn.read_bytes()).hexdigest()
        output_guid = generateGuid(output_md5, "MD5")
        job_info = self._format_bk_job_info(
            transform,
            file_key,
            start,
            end,
            output_fn.stat().st_size,
            output_md5,
            output_guid,
        )

        # Upload the merged histogram file and register in Bookkeeping
        self._upload_and_register(output_lfn, output_fn, job_info, output_guid)
        output_pfn = self._query_pfns([output_lfn], dummy=not gDoRealMerging)[output_lfn]
        # Delete the local merged histogram file
        output_fn.unlink()

        # Generate the metadata YAML file to be read by Monet
        metadata = {"LFN": output_lfn, "PFN": output_pfn, "refLFN": output_lfn, "refPFN": output_pfn}
        metadata = self._format_metadata(transform, request, metadata)
        metadata_dir = self.simdq_dir / str(request["RequestID"]).rjust(8, "0") / transform["Parameters"]["eventType"]
        secondary_app_name = self._get_application_name(transform, file_key, secondary=True)
        step_index = file_key.split("-")[0]
        metadata_fn = f"{step_index}-{secondary_app_name.upper()}HIST.yml"
        self._write_metadata_file(metadata_dir, metadata_fn, metadata)

        # Return the path to the metadata YAML file so it can be committed
        return metadata_dir / metadata_fn

    # Methods that perform queries with the clients ############################

    def _query_recent_transformations(self) -> list[dict]:
        """Get transformations that have finished testing but are not yet archived"""
        if gStandAlone:
            return []
        query = {
            "Type": self.allowed_types,
            "Status": self.allowed_statuses,
        }
        transformations = returnValueOrRaise(self.tf_client.getTransformations(query))
        self.log.info(f"Found {len(transformations)} recent transformations with status {self.allowed_statuses}")
        return transformations

    def _query_requests(self, request_ids: list[int]) -> dict[int, dict]:
        if gStandAlone:
            # Rather than empty list, return dicts with expected keys but blank values
            return {req_id: {key: "" for key in REQ_KEYS} for req_id in request_ids}
        requests = returnValueOrRaise(self.rp_client.getProductionRequest(request_ids))
        self.log.info(f"Found {len(requests)} requests")
        return requests

    def _query_input_lfns(self, transform_ids: list[int]) -> list[str]:
        if gStandAlone:
            return []
        query = {
            "Production": transform_ids,
            "FileType": self._available_hist_filetypes,
            "Visibility": "Y",
            "ReplicaFlag": "Yes",
        }
        ret = returnValueOrRaise(self.bk_client.getFilesWithMetadata(query))
        lfn_records = [dict(zip(ret["ParameterNames"], record)) for record in ret["Records"]]
        all_lfns = [x["FileName"] for x in lfn_records]
        self.log.info(f"Found {len(all_lfns)} LFNs from {len(transform_ids)} transformations")
        return all_lfns

    def _query_pfns(self, lfns: list[str], *, dummy: bool = False) -> dict[str, str]:
        if dummy or gStandAlone:
            # Assume the LFNs are at CERN
            return {lfn: "root://eoslhcb.cern.ch//eos/lhcb/grid/prod" + lfn for lfn in lfns}
        replicas = returnValueOrRaise(self.dm_client.getReplicas(lfns, diskOnly=True, getUrl=True, protocol="root"))
        if replicas["Failed"]:
            raise NotImplementedError(replicas)
        all_pfns = {lfn: list(pfns.values())[0] for lfn, pfns in replicas["Successful"].items()}
        self.log.info(f"Found {len(all_pfns)} PFNs from {len(lfns)} LFNs")
        return all_pfns

    def _query_lfns_exist(self, lfns: list[str], *, expected: str = "Failed") -> dict:
        if gStandAlone:
            return {expected: lfns}
        result = returnValueOrRaise(self.bk_client.getFileMetadata(lfns))
        msg = ", ".join([f"{state}: {len(result[state])} LFNs" for state in result])
        self.log.info(msg)
        return result

    # Methods that transform existing data #####################################

    @staticmethod
    def _group_lfns_by_prod_and_type(lfns: list[str]) -> dict[int, dict[str, list[str]]]:
        """Group a list of LFNs by their transform ID and step+filetype, as nested dicts, e.g.:

        .. code-block:: python

            {
                1234: {"1-GAUSSHIST" : ["/lhcb/MC/2011/GAUSSHIST/...", ...]},
                1235: {"1-BOOLEHIST" : ["/lhcb/MC/2011/BOOLEHIST/...", ...]},
            }

        :param lfns: Flat list of LFNs to sort
        :returns: Nested dicts where the outer key is transform ID, and the
                  inner key is <step>_<filetype>, and the inner values are lists
                  of LFNs
        """
        lfn_regex = "/".join(
            [
                r"/lhcb/MC/(?P<data_type>[\w\.\-]+)",
                r"(?P<filetype>[A-Z]+HIST)",
                r"(?P<transform_id>\d{8})",
                r"\d{4}",
                r"(?P<application>\w+)\_\d{8}\_\d{8}\_(?P<step>\d+)\.Hist\.root",
            ]
        )
        pattern = re.compile(lfn_regex)
        sorted_lfns = {}
        for lfn in lfns:
            if (match := re.match(pattern, lfn)) is None:
                raise NotImplementedError(f"{lfn=} does not match pattern {lfn_regex}")
            # Pull out the strings used as/in dict keys
            parent_transform = int(match.group("transform_id"))
            filetype = match.group("filetype")
            step_index = match.group("step")
            key = "-".join([step_index, filetype])
            # Append the LFN to the right nested dict
            if parent_transform not in sorted_lfns:
                sorted_lfns[parent_transform] = {key: []}
            if key not in sorted_lfns[parent_transform]:
                sorted_lfns[parent_transform][key] = []
            sorted_lfns[parent_transform][key] += [lfn]
        return sorted_lfns

    def _parse_workflow_xml(self, xml: str) -> XMLNode:
        workflow = self.xml_parser.parseString(xml)[0]
        if workflow.name != "Workflow":
            raise NotImplementedError(f"Expected an XML Workflow node, found {workflow}")
        return workflow

    def _get_transformation_parameters(self, workflow: XMLNode) -> dict:
        """Extract the <Parameter> XML tags from the transform Body, convert to a dict"""
        parameters = {child.attributes["name"]: child.children[0].value for child in workflow.childrens("Parameter")}
        return parameters

    def _get_transformation_steps(self, workflow: XMLNode) -> dict[str, dict]:
        """Extract the <Parameter> XML tags from each <StepInstance> in the
        transform Body, convert to dict where the keys are the values of the
        <name> tags, and the values are dicts of the <Parameter> tags
        """
        step_instances = workflow.childrens("StepInstance")
        steps = {
            step.childrens("name")[0].value: {
                child.attributes["name"]: child.children[0].value for child in step.childrens("Parameter")
            }
            for step in step_instances
        }
        return steps

    @staticmethod
    def _is_dict_string(string: str) -> bool:
        """Deduce whether a string encodes a python dict as YAML/JSON"""
        if not isinstance(string, str):
            return False
        try:
            return isinstance(yaml.safe_load(string), dict)
        except yaml.parser.ParserError:
            return False

    def _parse_dict_strings(self, data: dict) -> dict:
        """Convert "string dicts" to actual dicts"""
        return {key: yaml.safe_load(value) if self._is_dict_string(value) else value for key, value in data.items()}

    def _format_metadata(self, transform: dict, request: dict, extra: dict) -> dict:
        """Content of the metadata YAML file to for SimDQ"""
        req_dict = {key: request[key] for key in REQ_KEYS}
        req_dict = self._parse_dict_strings(req_dict)

        tfm_dict = {key: transform[key] for key in TFM_KEYS}
        tfm_dict = self._parse_dict_strings(tfm_dict)

        metadata = copy(req_dict)
        metadata.update(tfm_dict)
        metadata.update(extra)

        return metadata

    @staticmethod
    def _get_application_name(transform: dict, file_key: str, *, secondary: bool = False) -> str:
        """Deduce the application name with the correct capitalisation"""
        step_index, filetype = file_key.split("-")
        matching_steps = list(
            filter(
                lambda step_name: str(step_name).upper() == f"{filetype.removesuffix('HIST')}_{step_index}",
                transform["Steps"].keys(),
            )
        )
        assert len(matching_steps) == 1
        app_key = matching_steps[0]
        step = transform["Steps"][app_key]
        if secondary and step["optionsFormat"]:
            return step["optionsFormat"]
        return step["applicationName"]

    def _format_output_lfn(self, transform: dict, file_key: str, merge_id: int = 0) -> str:
        step_index, filetype = file_key.split("-")
        transform_id = int(transform["TransformationID"])
        parameters = transform["Parameters"]
        app = self._get_application_name(transform, file_key)
        output_lfn = "/".join(
            [
                "/lhcb",
                parameters["configName"],
                parameters["configVersion"],
                f"{filetype}MERGED",
                f"{transform_id:08d}",
                "0" * 4,
                f"{app}_{transform_id:08d}_{merge_id}_{step_index}.Hist.root",
            ]
        )
        return output_lfn

    def _format_bk_job_info(
        self,
        transform: dict,
        file_key: str,
        start: datetime,
        end: datetime,
        output_filesize: int,
        output_md5: str,
        output_guid: str,
    ) -> BookkeepingJobInfo:
        transform_id = int(transform["TransformationID"])
        parameters = transform["Parameters"]
        merge_info = transform["HistogramMerging"][file_key]
        input_lfns = merge_info["InputLFNs"]
        output_lfn = merge_info["OutputLFN"]
        filetype = file_key.rsplit("-", maxsplit=1)[-1]
        fn_time = start.strftime("%Y%m%d%H%M%S")
        job_info = BookkeepingJobInfo(
            ConfigName=parameters["configName"],
            ConfigVersion=parameters["configVersion"],
            Date=end.strftime("%Y-%m-%d"),
            Time=end.strftime("%H:%M:%S"),
        )
        if simulation_condition := parameters.get("BKCondition"):
            job_info.simulation_condition = simulation_condition
        job_info.typed_parameters = BookkeepingJobInfo.TypedParameters(
            CPUTIME=f"{(end - start).total_seconds():.0f}",
            ExecTime=f"{(end - start).total_seconds() * self.n_cpus:.0f}",
            NumberOfProcessors=f"{self.n_cpus}",
            Production=f"{transform_id}",
            Name=f"{transform_id:08d}_{fn_time}_1",
            JobStart=start.strftime("%Y-%m-%d %H:%M"),
            JobEnd=end.strftime("%Y-%m-%d %H:%M"),
            Location=DIRAC.siteName(),
            ProgramName=f"lb-conda {self.lbconda_env_name}",
            ProgramVersion=self.lbconda_env_version,
            StepID="1",
            NumberOfEvents=f"{len(input_lfns)}",
        )
        job_info.input_files = input_lfns
        job_info.output_files = [
            BookkeepingJobInfo.OutputFile(
                Name=output_lfn,
                TypeName=f"{filetype}MERGED",
                TypeVersion="1",
                EventTypeId=parameters["eventType"],
                EventStat=f"{len(input_lfns)}",
                FileSize=str(output_filesize),
                CreationDate=end.strftime("%Y-%m-%d %H:%M:%S"),
                MD5Sum=output_md5,
                Guid=output_guid,
            ),
        ]
        return job_info

    # Methods that write and commit files ######################################

    @staticmethod
    def _write_metadata_file(metadata_dir: Path, metadata_fn: str, metadata: dict):
        metadata_dir.mkdir(parents=True, exist_ok=True)
        with open(metadata_dir / metadata_fn, "w", encoding="utf-8") as metadata_file:
            yaml.dump(metadata, metadata_file)

    @timeThis
    def _write_merged_hist_file(self, input_pfns: list[str], output_fn: str) -> subprocess.CompletedProcess:
        with tempfile.NamedTemporaryFile(mode="wt") as tmpfile:
            tmpfile.write("\n".join(input_pfns))
            tmpfile.flush()
            lbconda_env = f"{self.lbconda_env_name}/{self.lbconda_env_version}"
            if gDoRealMerging:
                cmd = [
                    "lb-conda",
                    lbconda_env,
                    "hadd",
                    "-ff",
                    "-j",
                    str(self.n_cpus),
                    str(output_fn),
                    f"@{tmpfile.name}",
                ]
            else:
                cmd = ["touch", str(output_fn)]
            self.log.info("Running command:", shlex.join(cmd))
            ret = subprocess.run(cmd, capture_output=True, text=True)
            level = self.log.debug if ret.returncode == 0 else self.log.info
            level("stdout:", ret.stdout)
            level("stderr:", ret.stderr)
            ret.check_returncode()
            return ret

    def _upload_and_register(self, output_lfn, output_fn, job_info, output_guid):
        bk_xml = job_info.to_xml().decode()
        if gStandAlone or not gDoRealUpdate:
            return
        returnValueOrRaise(self.bk_client.sendXMLBookkeepingReport(bk_xml))
        res = returnValueOrRaise(self.dm_client.putAndRegister(output_lfn, str(output_fn), self.hist_se, output_guid))
        if res["Failed"]:
            raise NotImplementedError(res)
        if {"put", "register"} - set(res["Successful"].get(output_lfn, [])):
            raise NotImplementedError(res)

    def _checkout_metadata(self):
        if (self.simdq_dir / ".git").is_dir():
            repo = git.Repo(self.simdq_dir)
            with repo.git.custom_environment(GIT_SSH_COMMAND=f"ssh -i {self.ssh_key}"):
                repo.remotes.origin.fetch()
            return repo

        repo = git.Repo.init(self.simdq_dir)
        if gUseGitRepo:
            if "origin" in [r.name for r in repo.remotes]:
                origin = repo.remotes.origin
                origin.set_url(self.simdq_url)
            else:
                origin = repo.create_remote("origin", self.simdq_url)
            with repo.git.custom_environment(GIT_SSH_COMMAND=f"ssh -i {self.ssh_key}"):
                origin.fetch()
            remote_branch = getattr(origin.refs, self.simdq_branch)
            repo.create_head(self.simdq_branch, remote_branch).set_tracking_branch(remote_branch).checkout()
        return repo

    def _commit_and_push_metadata(self, new_files: list[Path], message: str):
        if gStandAlone or not (gUseGitRepo and gDoRealUpdate):
            return
        with self.simdq_repo.git.custom_environment(GIT_SSH_COMMAND=f"ssh -i {self.ssh_key}"):
            self.simdq_repo.remotes.origin.pull(rebase=True, autostash=True)
            self.simdq_repo.index.add(map(str, new_files))
            self.simdq_repo.index.commit(message, author=self.git_author, committer=self.git_author)
            self.simdq_repo.remotes.origin.push().raise_if_error()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC's Resources helper."""
import json

import LbPlatformUtils

from DIRAC import S_OK, S_ERROR, gLogger
import DIRAC.ConfigurationSystem.Client.Helpers.Resources

try:
    from json import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError

getQueues = DIRAC.ConfigurationSystem.Client.Helpers.Resources.getQueues
getDIRACPlatforms = DIRAC.ConfigurationSystem.Client.Helpers.Resources.getDIRACPlatforms
getCompatiblePlatforms = DIRAC.ConfigurationSystem.Client.Helpers.Resources.getCompatiblePlatforms

DEFAULT_CACHEPATH = "/cvmfs/lhcb.cern.ch/lib/var/lib/softmetadata/project-platforms.json"


def getDIRACPlatform(platform):
    """Returns list of compatible platforms.

    Used in JobDB.py instead of DIRAC.ConfigurationSystem.Client.Helpers.Resources.getDIRACPlatform

    :param str platform: a string (or a list with 1 string in)
                         representing a DIRAC platform, e.g. x86_64-centos7.avx2+fma
    :returns: S_ERROR or S_OK() with a list of DIRAC platforms that can run platform (e.g. slc6 can run on centos7)
    """

    # In JobDB.py this function is called with a list in input
    # In LHCb it should always be 1 and 1 only. If it's more there's an issue.
    if isinstance(platform, list):
        if len(platform) > 1:
            return S_ERROR("More than 1 platform specified for the job")
        platform = platform[0]

    if not platform or platform.lower() == "any":
        return S_OK([])

    return S_OK(LbPlatformUtils.compatible_platforms(platform))


def getPlatformForJob(workflow):
    """Looks inside the steps definition to find all requested CMTConfigs
    ("binary tag"), then translates it in a DIRAC platform.

    A binary tag is in the form of

    .. code-block:: none

      arch+microarch-osversion-gccversion-opt

    e.g.: x86_64+avx2+fma-centos7-gcc7-opt, x86_64-slc6-gcc49-opt

    We want to know what the worklow (the job) requires, so we need to "compose" the requested config
    and then get the minimum DIRAC platform that can run it.
    If, for example, the step1 and step2 respectively requires

    .. code-block:: none

      x86_64+avx2+fma-slc6-gcc7-opt
      x86_64-centos7-gcc62-opt

    then we conclude that to run this job we need

    .. code-block:: none

      x86_64+avx2+fma-centos7-gcc7-opt

    and so the DIRAC platform x86_64-centos7.avx2+fma

    :returns: a DIRAC platform (a string) or None
    """
    binaryTags = _findBinaryTags(workflow)

    if not binaryTags:
        gLogger.debug("Resources.getPlatformForJob: this job has no specific binary tag requested in any of its steps")
        return None

    return LbPlatformUtils.lowest_common_requirement(binaryTags)


def _findBinaryTags(wf):
    """developer function
    :returns: set of binary tags found in the workflow, each element is a
    frozenset of platform alternatives
    """
    binaryTags = set()
    for step_instance in wf.step_instances:
        systemConfig = step_instance.findParameter("SystemConfig")
        # If present, prefer the step defined system config
        if systemConfig and systemConfig.value.lower() != "any":
            binaryTags.add(frozenset([systemConfig.value]))
            continue

        # Else, if there is an application, query the SoftConfDB
        applicationName = step_instance.findParameter("applicationName")
        applicationVersion = step_instance.findParameter("applicationVersion")
        if not (applicationName and applicationVersion):
            continue

        platforms = _listPlatforms(applicationName.value, applicationVersion.value, DEFAULT_CACHEPATH)

        if platforms:
            binaryTags.add(frozenset(platforms))

    return binaryTags


def _listPlatforms(applicationName, applicationVersion, cachePath):
    """developer function
    :returns: set of binary tags found for a given application and version
    """
    applicationName = applicationName.upper()
    applicationVersion = applicationVersion.lower()
    platforms = None

    try:
        with open(cachePath) as fp:
            projectsMetadata = json.load(fp)
    except (FileNotFoundError, JSONDecodeError):
        gLogger.error("SoftConfDB JSON cache not found in", cachePath)
    else:
        try:
            platforms = projectsMetadata[applicationName][applicationVersion]
        except KeyError as e:
            gLogger.error(
                "Failed to find platform in cache for",
                f"{applicationName}/{applicationVersion} in {cachePath} ({e})",
            )

    return platforms



###############################################################################
# (c) Copyright 2023 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import importlib_metadata as metadata


def test_entrypoints():
    """Make sure all console_scripts defined by LHCbDIRAC are importable."""
    errors = []
    for ep in metadata.entry_points(group="console_scripts"):
        if ep.module.startswith("LHCbDIRAC"):
            try:
                ep.load()
            except ModuleNotFoundError as e:
                errors.append(str(e))
    assert not errors, errors



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Module to upload specified job output files according to the parameters
defined in the user workflow."""
import os

from DIRAC import S_OK, S_ERROR, gLogger, gConfig
from DIRAC.Core.Utilities import DEncode
from DIRAC.RequestManagementSystem.Client.Request import Operation
from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from LHCbDIRAC.DataManagementSystem.Client.DataUsageClient import DataUsageClient


class FileUsage(ModuleBase):
    #############################################################################

    def __init__(self, bkClient=None, dm=None):
        """Module initialization."""
        self.log = gLogger.getSubLogger("FileUsage")
        super().__init__(self.log, bkClientIn=bkClient, dm=dm)
        self.dataUsageClient = DataUsageClient()

    #############################################################################

    def _resolveInputVariables(self):
        """By convention the module parameters are resolved here."""
        super()._resolveInputVariables()

        dirDict = {}
        if self.inputDataList:
            for inputFile in self.inputDataList:
                strippedDir = os.path.join(os.path.dirname(inputFile), "")
                if not strippedDir:
                    self.log.error("File specified without directory! ", inputFile)
                else:
                    dirDict[strippedDir] = dirDict.setdefault(strippedDir, 0) + 1
            self.log.info("Popularity report = ", dirDict)
        else:
            self.log.info("No input data specified for this job")

        return S_OK(dirDict)

    #############################################################################

    def execute(
        self,
        production_id=None,
        prod_job_id=None,
        wms_job_id=None,
        workflowStatus=None,
        stepStatus=None,
        wf_commons=None,
        step_commons=None,
        step_id=None,
        step_number=None,
    ):
        """Main execution function."""

        try:
            super().execute(
                production_id,
                prod_job_id,
                wms_job_id,
                workflowStatus,
                stepStatus,
                wf_commons,
                step_commons,
                step_number,
                step_id,
            )

            result = self._resolveInputVariables()
            if not result["OK"]:
                self.log.error(result["Message"])
                return S_OK()
            dirDict = result["Value"]

            self.request.RequestName = "job_%d_request.xml" % self.jobID
            self.request.JobID = self.jobID
            self.request.SourceComponent = "Job_%d" % self.jobID

            if dirDict:
                result = self._reportFileUsage(dirDict)
                if not result["OK"]:
                    self.log.error(result["Message"])
                    return S_OK()
                self.log.info("Reporting input file usage successful!")
            else:
                self.log.info("No input data usage to report!")

            return S_OK("File Usage reported successfully")

        except Exception as e:  # pylint:disable=broad-except
            self.log.exception("Failure in FileUsage execute module", lException=e)
            return S_ERROR(str(e))

        finally:
            super().finalize()

    #############################################################################

    def _reportFileUsage(self, dirDict):
        """Send the data usage report (site,dirDict) where dirDict = {'Dataset':NumberOfHits}
        example: {'/lhcb/certification/test/ALLSTREAMS.DST/00000002/0000/': 1,
        '/lhcb/LHCb/Collision11/BHADRON.DST/00012957/0000/': 2}
        """
        self.log.verbose("Reporting input file usage:")
        for entry in dirDict:
            self.log.verbose(f"{entry}:{dirDict[entry]}")
        localSite = gConfig.getValue("/LocalSite/Site", "UNKNOWN")
        try:
            localSite = localSite.split(".")[1]
        except BaseException:
            pass
        self.log.verbose(f"Using Site Name: {localSite}")

        if self._enableModule():
            usageStatus = self.dataUsageClient.sendDataUsageReport(localSite, dirDict)
            if not usageStatus["OK"]:
                self.log.error("Could not send data usage report, preparing a DISET failover request object")
                self.log.verbose(usageStatus["rpcStub"])
                forwardDISETOp = Operation()
                forwardDISETOp.Type = "ForwardDISET"
                forwardDISETOp.Arguments = DEncode.encode(usageStatus["rpcStub"])
                self.request.addOperation(forwardDISETOp)
                self.workflow_commons["Request"] = self.request
        else:
            self.log.info(f"Would have attempted to report {dirDict} at {localSite}")
            return S_OK()

        return S_OK()


# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Module holding ProductionRequestClient class."""
from DIRAC.Core.Base.Client import Client, createClient


@createClient("ProductionManagement/ProductionRequest")
class ProductionRequestClient(Client):
    """This class expose the methods of the Production Request Service."""

    def __init__(self, url=None, **kwargs):
        """c'tor.

        :param str url: can specify a specific URL
        """
        super().__init__(**kwargs)
        self.setServer("ProductionManagement/ProductionRequest")
        if url:
            self.setServer(url)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""dirac-production-runjoblocal.

Module created to run failed jobs locally on a CVMFS-configured machine.
It creates the necessary environment, downloads the necessary files, modifies the necessary
files and runs the job

Usage:
  dirac-production-runjoblocal (job ID) (Data imput mode) -  No parenthesis
"""
import sys
import os
import shutil
import ssl

from urllib.request import urlopen as url_library_urlopen  # pylint: disable=no-name-in-module,import-error
from urllib.error import URLError as url_library_URLError  # pylint: disable=no-name-in-module,import-error

from DIRAC import S_OK
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.Core.Base.Script import Script


def __runSystemDefaults(jobID=None):
    """Creates the environment for running the job and returns the path for the
    other functions."""
    tempdir = "LHCbjob" + str(jobID) + "temp"
    os.environ["VO_LHCB_SW_DIR"] = "/cvmfs/lhcb.cern.ch"
    mkDir(tempdir)

    basepath = os.getcwd()
    return basepath + os.path.sep + tempdir + os.path.sep


def __downloadJobDescriptionXML(jobID, basepath):
    """Downloads the jobDescription.xml file into the temporary directory
    created."""
    from DIRAC.Interfaces.API.Dirac import Dirac

    jdXML = Dirac()
    jdXML.getInputSandbox(jobID, basepath)


def __modifyJobDescription(jobID, basepath, downloadinputdata):
    """Modifies the jobDescription.xml to, instead of DownloadInputData, it uses
    InputDataByProtocol."""
    if not downloadinputdata:
        from xml.etree import ElementTree as et

        archive = et.parse(basepath + "InputSandbox" + str(jobID) + os.path.sep + "jobDescription.xml")
        for element in archive.iter():
            if element.text == "DIRAC.WorkloadManagementSystem.Client.DownloadInputData":
                element.text = "DIRAC.WorkloadManagementSystem.Client.InputDataByProtocol"
                archive.write(basepath + "InputSandbox" + str(jobID) + os.path.sep + "jobDescription.xml")
                return S_OK("Job parameter changed from DownloadInputData to InputDataByProtocol.")


def __downloadPilotScripts(basepath):
    """
    Downloads the scripts necessary to configure the pilot
    """
    context = ssl._create_unverified_context()
    for fileName in [
        "dirac-pilot.py",
        "dirac-install.py",
        "pilotCommands.py",
        "pilotTools",
        "MessageSender",
        "PilotLogger.py",
        "PilotLoggerTools.py",
    ]:
        remoteFile = url_library_urlopen(
            os.path.join("https://raw.githubusercontent.com/DIRACGrid/Pilot/master/Pilot/", fileName),
            timeout=10,
            context=context,
        )
        with open(fileName, "wb") as localFile:
            localFile.write(remoteFile.read())

    remoteFile = url_library_urlopen(
        os.path.join("https://gitlab.cern.ch/lhcb-dirac/LHCbPilot/-/raw/master/LHCbPilot/LHCbPilotCommands.py"),
        timeout=10,
        context=context,
    )
    with open("LHCbPilotCommands.py", "wb") as localFile:
        localFile.write(remoteFile.read())
    remoteFile = url_library_urlopen(
        os.path.join("https://gitlab.cern.ch/lhcb-dirac/LHCbPilot/-/raw/master/LHCbPilot/LHCbPilotTools.py"),
        timeout=10,
        context=context,
    )
    with open("LHCbPilotTools.py", "wb") as localFile:
        localFile.write(remoteFile.read())


def __configurePilot(basepath):
    """Configures the pilot."""
    pilotCmd = "dirac-pilot.py -S LHCb-Production -l LHCb "
    pilotCmd += "-C dips://lhcb-conf-dirac.cern.ch:9135/Configuration/Server "
    pilotCmd += "-N ce.debug.ch -Q default -n DIRAC.JobDebugger.cern -M 1 "
    pilotCmd += "-E LHCbPilot "
    pilotCmd += "-X LHCbConfigureBasics,LHCbConfigureSite,LHCbConfigureArchitecture,LHCbConfigureCPURequirements -dd"
    out = os.system("python " + basepath + pilotCmd)
    if not out:
        directory = os.path.expanduser("~") + os.path.sep
        os.rename(directory + ".dirac.cfg", directory + ".dirac.cfg.old")
        shutil.copyfile(directory + "pilot.cfg", directory + ".dirac.cfg")
        return S_OK("Pilot successfully configured.")


#   else:
#     some DErrno message


def __runJobLocally(jobID, basepath):
    """Runs the job!"""
    from LHCbDIRAC.Interfaces.API.LHCbJob import LHCbJob

    localJob = LHCbJob(basepath + "InputSandbox" + str(jobID) + os.path.sep + "jobDescription.xml")
    localJob.setInputSandbox(os.getcwd() + "pilot.cfg")
    localJob.setConfigArgs(os.getcwd() + "pilot.cfg")
    os.chdir(basepath)
    localJob.runLocal()


@Script()
def main():
    Script.registerSwitch("D:", "Download=", "Defines data acquisition as DownloadInputData")
    Script.registerSwitch("P:", "Protocol=", "Defines data acquisition as InputDataByProtocol")
    Script.parseCommandLine(ignoreErrors=False)

    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "\nUsage:",
                "dirac-production-runjoblocal [Data imput mode] [job ID]" "\nArguments:",
                "  Download (Job ID): Defines data acquisition as DownloadInputData",
                "  Protocol (Job ID): Defines data acquisition as InputDataByProtocol\n",
            ]
        )
    )

    _downloadinputdata = False
    _jobID = None

    for switch in Script.getUnprocessedSwitches():
        if switch[0] in ("D", "Download"):
            _downloadinputdata = True
            _jobID = switch[1]
        if switch[0] in ("P", "Protocol"):
            _downloadinputdata = False
            _jobID = switch[1]

    usedDir = os.path.expanduser("~") + os.path.sep
    try:
        _path = __runSystemDefaults(_jobID)

        __downloadJobDescriptionXML(_jobID, _path)

        __modifyJobDescription(_jobID, _path, _downloadinputdata)

        __downloadPilotScripts(_path)

        __configurePilot(_path)

        __runJobLocally(_jobID, _path)

    finally:
        os.chdir(usedDir)
        os.rename(usedDir + ".dirac.cfg.old", usedDir + ".dirac.cfg")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Add files to a transformation."""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    import os
    import DIRAC

    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerFileSwitches()

    Script.registerSwitch("", "NoRunInfo", "   Use if no run information is required")
    Script.registerSwitch("", "ResetPresentFiles", "   Reset files Unused if they are present in the transformation")
    Script.registerSwitch("", "Chown=", "   Give user/group for chown of the directories of files in the FC")

    Script.parseCommandLine(ignoreErrors=True)
    from DIRAC import gLogger
    from DIRAC.TransformationSystem.Utilities.ScriptUtilities import getTransformations

    Script.setUsageMessage(
        "\n".join(
            [
                __doc__,
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ...",
            ]
        )
    )

    runInfo = True
    userGroup = None
    resetPresentFiles = False

    switches = Script.getUnprocessedSwitches()
    for opt, val in switches:
        if opt == "NoRunInfo":
            runInfo = False
        elif opt == "ResetPresentFiles":
            resetPresentFiles = True
        elif opt == "Chown":
            userGroup = val.split("/")
            if len(userGroup) != 2 or not userGroup[1].startswith("lhcb_"):
                gLogger.fatal("Wrong user/group")
                DIRAC.exit(2)

    if userGroup:
        from DIRAC.Core.Security.ProxyInfo import getProxyInfo

        res = getProxyInfo()
        if not res["OK"]:
            gLogger.fatal("Can't get proxy info", res["Message"])
            DIRAC.exit(1)
        properties = res["Value"].get("groupProperties", [])
        if "FileCatalogManagement" not in properties:
            gLogger.error("You need to use a proxy from a group with FileCatalogManagement")
            DIRAC.exit(5)

    from LHCbDIRAC.DataManagementSystem.Utilities.FCUtilities import chown
    from LHCbDIRAC.TransformationSystem.Utilities.PluginUtilities import addFilesToTransformation

    transList = getTransformations(Script.getPositionalArgs())
    if not transList:
        DIRAC.exit(1)

    requestedLFNs = dmScript.getOption("LFNs")
    if not requestedLFNs:
        gLogger.always("No files to add")
        DIRAC.exit(0)

    if userGroup:
        directories = {os.path.dirname(lfn) for lfn in requestedLFNs}
        res = chown(directories, user=userGroup[0], group=userGroup[1])
        if not res["OK"]:
            gLogger.fatal("Error changing ownership", res["Message"])
            DIRAC.exit(3)
        gLogger.notice("Successfully changed owner/group for %d directories" % res["Value"])

    rc = 0
    for transID in transList:
        res = addFilesToTransformation(transID, requestedLFNs, addRunInfo=runInfo, resetPresentFiles=resetPresentFiles)
        if res["OK"]:
            added = list(res["Value"]["Successful"].values()).count("Added")
            if added:
                gLogger.always("Successfully added %d files to transformation %d" % (added, transID))
            else:
                gLogger.always("No files added to transformation", str(transID))
            reset = list(res["Value"]["Successful"].values()).count("Reset")
            if reset:
                gLogger.always("Successfully reset %d files Unused in transformation %d" % (reset, transID))
            if res["Value"]["Failed"]:
                from collections import defaultdict

                errors = defaultdict(int)
                for error in res["Value"]["Failed"].values():
                    errors[error] += 1
                for error, count in errors.items():
                    gLogger.always("Failed to add %d files to transformation %d:" % (count, transID), error)
                rc = 1
        else:
            gLogger.always(
                "Failed to add %d files to transformation %d" % (len(requestedLFNs), transID), res["Message"]
            )
            rc = 2
    DIRAC.exit(rc)


if __name__ == "__main__":
    main()
